import numpy as np
import pandas as pd
import pytest

from featuretools.primitives import NumberOfWordsInQuotes
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestNumberOfWordsInQuotes(PrimitiveTestBase):
    primitive = NumberOfWordsInQuotes

    def test_regular_double_quotes_input(self):
        x = pd.Series(
            [
                'Yes "    "',
                '"Hello this is a test"',
                '"Yes" "   "',
                "",
                '"Python, java prolog"',
                '"Python, java prolog" three words here "binary search algorithm"',
                '"Diffie-Hellman key exchange"',
                '"user@email.com"',
                '"https://alteryx.com"',
                '"100,000"',
                '"This Borderlands game here"" is the perfect conclusion to the ""Borderlands 3"" line, which focuses on the fans ""favorite character and gives the players the opportunity to close for a long time some very important questions about\'s character and the memorable scenery with which the players interact.',
            ],
        )
        expected = pd.Series([0, 5, 1, 0, 3, 6, 3, 1, 1, 1, 6], dtype="Int64")
        actual = self.primitive("double").get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_captures_regular_single_quotes(self):
        x = pd.Series(
            [
                "'Hello this is a test'",
                "'Python, Java Prolog'",
                "'Python, Java Prolog' three words here 'three words here'",
                "'Diffie-Hellman key exchange'",
                "'user@email.com'",
                "'https://alteryx.com'",
                "'there's where's here's' word 'word'",
                "'100,000'",
            ],
        )
        expected = pd.Series([5, 3, 6, 3, 1, 1, 4, 1], dtype="Int64")
        actual = self.primitive("single").get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_captures_both_single_and_double_quotes(self):
        x = pd.Series(
            [
                "'test test test test' three words here \"test test test!\"",
            ],
        )
        expected = pd.Series([7], dtype="Int64")
        actual = self.primitive().get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_unicode_input(self):
        x = pd.Series(
            [
                '"Ángel"',
                '"Ángel" word word',
            ],
        )
        expected = pd.Series([1, 1], dtype="Int64")
        actual = self.primitive().get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_multiline(self):
        x = pd.Series(
            [
                "'Yes\n, this is me'",
            ],
        )
        expected = pd.Series([4], dtype="Int64")
        actual = self.primitive().get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_raises_error_invalid_args(self):
        error_msg = (
            "NULL is not a valid quote_type. Specify 'both', 'single', or 'double'"
        )
        with pytest.raises(
            ValueError,
            match=error_msg,
        ):
            self.primitive(quote_type="NULL")

    def test_null(self):
        x = pd.Series([np.nan, pd.NA, None, '"test"'])
        actual = self.primitive().get_function()(x)
        expected = pd.Series([pd.NA, pd.NA, pd.NA, 1.0], dtype="Int64")
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



from featuretools import Feature, calculate_feature_matrix
from featuretools.primitives import (
    AddNumeric,
    AddNumericScalar,
    Count,
    DivideByFeature,
    DivideNumeric,
    DivideNumericScalar,
    Equal,
    EqualScalar,
    GreaterThan,
    GreaterThanEqualTo,
    GreaterThanEqualToScalar,
    GreaterThanScalar,
    LessThan,
    LessThanEqualTo,
    LessThanEqualToScalar,
    LessThanScalar,
    ModuloByFeature,
    ModuloNumeric,
    ModuloNumericScalar,
    MultiplyNumeric,
    MultiplyNumericScalar,
    Negate,
    NotEqual,
    NotEqualScalar,
    ScalarSubtractNumericFeature,
    SubtractNumeric,
    SubtractNumericScalar,
    Sum,
)
from featuretools.tests.testing_utils import to_pandas


def test_overrides(es):
    value = Feature(es["log"].ww["value"])
    value2 = Feature(es["log"].ww["value_2"])

    feats = [
        AddNumeric,
        SubtractNumeric,
        MultiplyNumeric,
        DivideNumeric,
        ModuloNumeric,
        GreaterThan,
        LessThan,
        Equal,
        NotEqual,
        GreaterThanEqualTo,
        LessThanEqualTo,
    ]
    assert Feature(value, primitive=Negate).unique_name() == (-value).unique_name()

    compares = [(value, value), (value, value2)]
    overrides = [
        value + value,
        value - value,
        value * value,
        value / value,
        value % value,
        value > value,
        value < value,
        value == value,
        value != value,
        value >= value,
        value <= value,
        value + value2,
        value - value2,
        value * value2,
        value / value2,
        value % value2,
        value > value2,
        value < value2,
        value == value2,
        value != value2,
        value >= value2,
        value <= value2,
    ]

    for left, right in compares:
        for feat in feats:
            f = Feature([left, right], primitive=feat)
            o = overrides.pop(0)
            assert o.unique_name() == f.unique_name()


def test_override_boolean(es):
    count = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )
    count_lo = Feature(count, primitive=GreaterThanScalar(1))
    count_hi = Feature(count, primitive=LessThanScalar(10))

    to_test = [[True, True, True], [True, True, False], [False, False, True]]

    features = []
    features.append(count_lo.OR(count_hi))
    features.append(count_lo.AND(count_hi))
    features.append(~(count_lo.AND(count_hi)))

    df = calculate_feature_matrix(
        entityset=es,
        features=features,
        instance_ids=[0, 1, 2],
    )
    df = to_pandas(df, index="id", sort_index=True)
    for i, test in enumerate(to_test):
        v = df[features[i].get_name()].tolist()
        assert v == test


def test_scalar_overrides(es):
    value = Feature(es["log"].ww["value"])

    feats = [
        AddNumericScalar,
        SubtractNumericScalar,
        MultiplyNumericScalar,
        DivideNumericScalar,
        ModuloNumericScalar,
        GreaterThanScalar,
        LessThanScalar,
        EqualScalar,
        NotEqualScalar,
        GreaterThanEqualToScalar,
        LessThanEqualToScalar,
    ]

    overrides = [
        value + 2,
        value - 2,
        value * 2,
        value / 2,
        value % 2,
        value > 2,
        value < 2,
        value == 2,
        value != 2,
        value >= 2,
        value <= 2,
    ]

    for feat in feats:
        f = Feature(value, primitive=feat(2))
        o = overrides.pop(0)
        assert o.unique_name() == f.unique_name()

    value2 = Feature(es["log"].ww["value_2"])

    reverse_feats = [
        AddNumericScalar,
        ScalarSubtractNumericFeature,
        MultiplyNumericScalar,
        DivideByFeature,
        ModuloByFeature,
        GreaterThanScalar,
        LessThanScalar,
        EqualScalar,
        NotEqualScalar,
        GreaterThanEqualToScalar,
        LessThanEqualToScalar,
    ]
    reverse_overrides = [
        2 + value2,
        2 - value2,
        2 * value2,
        2 / value2,
        2 % value2,
        2 < value2,
        2 > value2,
        2 == value2,
        2 != value2,
        2 <= value2,
        2 >= value2,
    ]
    for feat in reverse_feats:
        f = Feature(value2, primitive=feat(2))
        o = reverse_overrides.pop(0)
        assert o.unique_name() == f.unique_name()


def test_override_cmp_from_column(es):
    count_lo = Feature(es["log"].ww["value"]) > 1

    to_test = [False, True, True]

    features = [count_lo]

    df = to_pandas(
        calculate_feature_matrix(
            entityset=es,
            features=features,
            instance_ids=[0, 1, 2],
        ),
        index="id",
        sort_index=True,
    )
    v = df[count_lo.get_name()].tolist()
    for i, test in enumerate(to_test):
        assert v[i] == test


def test_override_cmp(es):
    count = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )
    _sum = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        primitive=Sum,
    )
    gt_lo = count > 1
    gt_other = count > _sum
    ge_lo = count >= 1
    ge_other = count >= _sum
    lt_hi = count < 10
    lt_other = count < _sum
    le_hi = count <= 10
    le_other = count <= _sum
    ne_lo = count != 1
    ne_other = count != _sum

    to_test = [
        [True, True, False],
        [False, False, True],
        [True, True, True],
        [False, False, True],
        [True, True, True],
        [True, True, False],
        [True, True, True],
        [True, True, False],
    ]
    features = [
        gt_lo,
        gt_other,
        ge_lo,
        ge_other,
        lt_hi,
        lt_other,
        le_hi,
        le_other,
        ne_lo,
        ne_other,
    ]

    df = calculate_feature_matrix(
        entityset=es,
        features=features,
        instance_ids=[0, 1, 2],
    )
    df = to_pandas(df, index="id", sort_index=True)
    for i, test in enumerate(to_test):
        v = df[features[i].get_name()].tolist()
        assert v == test



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Integer

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive


class CountOutsideNthSTD(AggregationPrimitive):
    """Determines the number of observations that lie outside
        the first N standard deviations.

    Args:
        n (float): Number of standard deviations. Default is 1

    Examples:
        >>> count_outside_nth_std = CountOutsideNthSTD(n=1.5)
        >>> count_outside_nth_std([1, 10, 15, 20, 100])
        1
    """

    name = "count_outside_nth_std"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def __init__(self, n=1):
        if n < 0:
            raise ValueError("n must be a positive number")

        self.n = n

    def get_function(self):
        def count_outside_nth_std(x):
            cond = np.abs(x - np.mean(x)) > np.std(x) * self.n
            return cond.sum()

        return count_outside_nth_std



import re

from featuretools.entityset.relationship import RelationshipPath


def feature_with_name(features, name):
    for f in features:
        if f.get_name() == name:
            return True
    return False


def number_of_features_with_name_like(features, pattern):
    """Returns number of features with names that match the provided regex pattern"""
    pattern = re.compile(re.escape(pattern))
    names = [f.get_name() for f in features]
    return len([name for name in names if pattern.search(name)])


def backward_path(es, dataframe_ids):
    """
    Create a backward RelationshipPath through the given dataframes. Assumes only
    one such path is possible.
    """

    def _get_relationship(child, parent):
        return next(
            r
            for r in es.get_forward_relationships(child)
            if r._parent_dataframe_name == parent
        )

    relationships = [
        _get_relationship(child, parent)
        for parent, child in zip(dataframe_ids[:-1], dataframe_ids[1:])
    ]

    return RelationshipPath([(False, r) for r in relationships])


def forward_path(es, dataframe_ids):
    """
    Create a forward RelationshipPath through the given dataframes. Assumes only
    one such path is possible.
    """

    def _get_relationship(child, parent):
        return next(
            r
            for r in es.get_forward_relationships(child)
            if r._parent_dataframe_name == parent
        )

    relationships = [
        _get_relationship(child, parent)
        for child, parent in zip(dataframe_ids[:-1], dataframe_ids[1:])
    ]

    return RelationshipPath([(True, r) for r in relationships])


def check_rename(feat, new_name, new_names):
    copy_feat = feat.rename(new_name)
    assert feat.unique_name() != copy_feat.unique_name()
    assert feat.get_name() != copy_feat.get_name()
    assert (
        feat.base_features[0].generate_name()
        == copy_feat.base_features[0].generate_name()
    )
    assert feat.dataframe_name == copy_feat.dataframe_name
    assert feat.get_feature_names() != copy_feat.get_feature_names()
    check_names(copy_feat, new_name, new_names)


def check_names(feat, new_name, new_names):
    assert feat.get_name() == new_name
    assert feat.get_feature_names() == new_names



from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class SubtractNumericScalar(TransformPrimitive):
    """Subtracts a scalar from each element in the list.

    Description:
        Given a list of numeric values and a scalar, subtract
        the given scalar from each value in the list.

    Examples:
        >>> subtract_numeric_scalar = SubtractNumericScalar(value=2)
        >>> subtract_numeric_scalar([3, 1, 2]).tolist()
        [1, -1, 0]
    """

    name = "subtract_numeric_scalar"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]

    def __init__(self, value=0):
        self.value = value
        self.description_template = "the result of {{}} minus {}".format(self.value)

    def get_function(self):
        def subtract_scalar(vals):
            return vals - self.value

        return subtract_scalar

    def generate_name(self, base_feature_names):
        return "%s - %s" % (base_feature_names[0], str(self.value))



import json
from unittest.mock import patch

import pytest
from woodwork.logical_types import Boolean, Double

from featuretools.feature_discovery.feature_discovery import (
    generate_features_from_primitives,
    schema_to_features,
)
from featuretools.feature_discovery.FeatureCollection import FeatureCollection
from featuretools.feature_discovery.LiteFeature import LiteFeature
from featuretools.primitives import (
    Absolute,
    AddNumeric,
    DivideNumeric,
    Lag,
    MultiplyNumeric,
)
from featuretools.tests.feature_discovery.test_feature_discovery import (
    MultiOutputPrimitiveForTest,
)
from featuretools.tests.testing_utils.generate_fake_dataframe import (
    generate_fake_dataframe,
)


def test_feature_type_equality():
    f1 = LiteFeature("f1", Double)
    f2 = LiteFeature("f2", Double)

    # Add Numeric is Commutative, so should all be equal
    f3 = LiteFeature(
        name="Column 1",
        primitive=AddNumeric(),
        logical_type=Double,
        base_features=[f1, f2],
    )

    f4 = LiteFeature(
        name="Column 10",
        primitive=AddNumeric(),
        logical_type=Double,
        base_features=[f1, f2],
    )

    f5 = LiteFeature(
        name="Column 20",
        primitive=AddNumeric(),
        logical_type=Double,
        base_features=[f2, f1],
    )

    assert f3 == f4 == f5

    # Divide Numeric is not Commutative, so should not be equal
    f6 = LiteFeature(
        name="Column 1",
        primitive=DivideNumeric(),
        logical_type=Double,
        base_features=[f1, f2],
    )

    f7 = LiteFeature(
        name="Column 1",
        primitive=DivideNumeric(),
        logical_type=Double,
        base_features=[f2, f1],
    )

    assert f6 != f7


def test_feature_type_assertions():
    with pytest.raises(
        ValueError,
        match="there must be base features if given a primitive",
    ):
        LiteFeature(
            name="Column 1",
            primitive=AddNumeric(),
            logical_type=Double,
        )


@patch.object(LiteFeature, "_generate_hash", lambda x: x.name)
@patch(
    "featuretools.feature_discovery.LiteFeature.hash_primitive",
    lambda x: (x.name, None),
)
def test_feature_to_dict():
    f1 = LiteFeature("f1", Double)
    f2 = LiteFeature("f2", Double)
    f = LiteFeature(
        name="Column 1",
        primitive=AddNumeric(),
        base_features=[f1, f2],
    )

    expected = {
        "name": "Column 1",
        "logical_type": None,
        "tags": ["numeric"],
        "primitive": "add_numeric",
        "base_features": ["f1", "f2"],
        "df_id": None,
        "id": "Column 1",
        "related_features": [],
        "idx": 0,
    }

    actual = f.to_dict()
    json_str = json.dumps(actual)
    assert actual == expected
    assert json.dumps(expected) == json_str


def test_feature_hash():
    bf1 = LiteFeature("bf", Double)
    bf2 = LiteFeature("bf", Double, df_id="df")

    p1 = Lag(periods=1)
    p2 = Lag(periods=2)
    f1 = LiteFeature(
        primitive=p1,
        logical_type=Double,
        base_features=[bf1],
    )

    f2 = LiteFeature(
        primitive=p2,
        logical_type=Double,
        base_features=[bf1],
    )

    f3 = LiteFeature(
        primitive=p2,
        logical_type=Double,
        base_features=[bf1],
    )

    f4 = LiteFeature(
        primitive=p1,
        logical_type=Double,
        base_features=[bf2],
    )

    # TODO(dreed): ensure ID is parquet and arrow acceptable, length and starting character might be problematic

    assert f1 != f2
    assert f2 == f3
    assert f1 != f4


def test_feature_forced_name():
    bf = LiteFeature("bf", Double)

    p1 = Lag(periods=1)
    f1 = LiteFeature(
        name="target_delay_1",
        primitive=p1,
        logical_type=Double,
        base_features=[bf],
    )
    assert f1.name == "target_delay_1"


@patch.object(LiteFeature, "_generate_hash", lambda x: x.name)
@patch(
    "featuretools.feature_discovery.FeatureCollection.hash_primitive",
    lambda x: (x.name, None),
)
@patch(
    "featuretools.feature_discovery.LiteFeature.hash_primitive",
    lambda x: (x.name, None),
)
def test_feature_collection_to_dict():
    f1 = LiteFeature("f1", Double)
    f2 = LiteFeature("f2", Double)
    f3 = LiteFeature(
        name="Column 1",
        primitive=AddNumeric(),
        base_features=[f1, f2],
    )

    fc = FeatureCollection([f3])

    expected = {
        "primitives": {
            "add_numeric": None,
        },
        "feature_ids": ["Column 1"],
        "all_features": {
            "Column 1": {
                "name": "Column 1",
                "logical_type": None,
                "tags": ["numeric"],
                "primitive": "add_numeric",
                "base_features": ["f1", "f2"],
                "df_id": None,
                "id": "Column 1",
                "related_features": [],
                "idx": 0,
            },
            "f1": {
                "name": "f1",
                "logical_type": "Double",
                "tags": ["numeric"],
                "primitive": None,
                "base_features": [],
                "df_id": None,
                "id": "f1",
                "related_features": [],
                "idx": 0,
            },
            "f2": {
                "name": "f2",
                "logical_type": "Double",
                "tags": ["numeric"],
                "primitive": None,
                "base_features": [],
                "df_id": None,
                "id": "f2",
                "related_features": [],
                "idx": 0,
            },
        },
    }

    actual = fc.to_dict()
    assert actual == expected
    assert json.dumps(expected, sort_keys=True) == json.dumps(actual, sort_keys=True)


@patch.object(LiteFeature, "_generate_hash", lambda x: x.name)
def test_feature_collection_from_dict():
    f1 = LiteFeature("f1", Double)
    f2 = LiteFeature("f2", Double)
    f3 = LiteFeature(
        name="Column 1",
        primitive=AddNumeric(),
        base_features=[f1, f2],
    )

    expected = FeatureCollection([f3])

    input_dict = {
        "primitives": {
            "009da67f0a1430630c4a419c84aac270ec62337ab20c080e4495272950fd03b3": {
                "type": "AddNumeric",
                "module": "featuretools.primitives.standard.transform.binary.add_numeric",
                "arguments": {},
            },
        },
        "feature_ids": ["Column 1"],
        "all_features": {
            "f2": {
                "name": "f2",
                "logical_type": "Double",
                "tags": ["numeric"],
                "primitive": None,
                "base_features": [],
                "df_id": None,
                "id": "f2",
                "related_features": [],
                "idx": 0,
            },
            "f1": {
                "name": "f1",
                "logical_type": "Double",
                "tags": ["numeric"],
                "primitive": None,
                "base_features": [],
                "df_id": None,
                "id": "f1",
                "related_features": [],
                "idx": 0,
            },
            "Column 1": {
                "name": "Column 1",
                "logical_type": None,
                "tags": ["numeric"],
                "primitive": "009da67f0a1430630c4a419c84aac270ec62337ab20c080e4495272950fd03b3",
                "base_features": ["f1", "f2"],
                "df_id": None,
                "id": "Column 1",
                "related_features": [],
                "idx": 0,
            },
        },
    }

    actual = FeatureCollection.from_dict(input_dict)

    assert actual == expected


@patch.object(LiteFeature, "__lt__", lambda x, y: x.name < y.name)
def test_feature_collection_serialization_roundtrip():
    col_defs = [
        ("idx", "Integer", {"index"}),
        ("t_idx", "Datetime", {"time_index"}),
        ("f_1", "Double"),
        ("f_2", "Double"),
        ("f_3", "Categorical"),
        ("f_4", "Boolean"),
        ("f_5", "NaturalLanguage"),
    ]

    df = generate_fake_dataframe(
        col_defs=col_defs,
    )

    origin_features = schema_to_features(df.ww.schema)
    features = generate_features_from_primitives(
        origin_features,
        [Absolute, MultiplyNumeric, MultiOutputPrimitiveForTest],
    )

    features = generate_features_from_primitives(features, [Lag])

    assert set([x.name for x in features]) == set(
        [
            "idx",
            "t_idx",
            "f_1",
            "f_2",
            "f_3",
            "f_4",
            "f_5",
            "ABSOLUTE(f_1)",
            "ABSOLUTE(f_2)",
            "f_1 * f_2",
            "TEST_MO(f_5)[0]",
            "TEST_MO(f_5)[1]",
            "LAG(f_1, t_idx)",
            "LAG(f_2, t_idx)",
            "LAG(f_3, t_idx)",
            "LAG(f_4, t_idx)",
            "LAG(ABSOLUTE(f_1), t_idx)",
            "LAG(ABSOLUTE(f_2), t_idx)",
            "LAG(f_1 * f_2, t_idx)",
            "LAG(TEST_MO(f_5)[1], t_idx)",
            "LAG(TEST_MO(f_5)[0], t_idx)",
        ],
    )
    fc = FeatureCollection(features=features)
    fc_dict = fc.to_dict()

    fc_json = json.dumps(fc_dict)

    fc2_dict = json.loads(fc_json)

    fc2 = FeatureCollection.from_dict(fc2_dict)

    assert fc == fc2
    lsa_features = [x for x in fc2.all_features if x.get_primitive_name() == "test_mo"]
    assert len(lsa_features[0].related_features) == 1


def test_lite_feature_assertions():
    f1 = LiteFeature(name="f1", logical_type=Double)
    f2 = LiteFeature(name="f1", logical_type=Double, df_id="df1")

    assert f1 != f2

    with pytest.raises(
        TypeError,
        match="Name must be given if origin feature",
    ):
        LiteFeature(logical_type=Double)

    with pytest.raises(
        TypeError,
        match="Logical Type must be given if origin feature",
    ):
        LiteFeature(name="f1")

    with pytest.raises(
        ValueError,
        match="primitive input must be of type PrimitiveBase",
    ):
        LiteFeature(name="f3", primitive="AddNumeric", base_features=[f1, f2])

    f = LiteFeature("f4", logical_type=Double)
    with pytest.raises(AttributeError, match="name is immutable"):
        f.name = "new name"

    with pytest.raises(ValueError, match="only used on multioutput features"):
        f.non_indexed_name

    with pytest.raises(AttributeError, match="logical_type is immutable"):
        f.logical_type = Boolean

    with pytest.raises(AttributeError, match="tags is immutable"):
        f.tags = {"other"}

    with pytest.raises(AttributeError, match="primitive is immutable"):
        f.primitive = AddNumeric

    with pytest.raises(AttributeError, match="base_features are immutable"):
        f.base_features = [f1]

    with pytest.raises(AttributeError, match="df_id is immutable"):
        f.df_id = "df_id"

    with pytest.raises(AttributeError, match="id is immutable"):
        f.id = "id"

    with pytest.raises(AttributeError, match="n_output_features is immutable"):
        f.n_output_features = "n_output_features"

    with pytest.raises(AttributeError, match="depth is immutable"):
        f.depth = "depth"

    with pytest.raises(AttributeError, match="idx is immutable"):
        f.idx = "idx"


def test_lite_feature_to_column_schema():
    f1 = LiteFeature(name="f1", logical_type=Double, tags={"index", "numeric"})

    column_schema = f1.column_schema

    assert column_schema.is_numeric
    assert isinstance(column_schema.logical_type, Double)
    assert column_schema.semantic_tags == {"index", "numeric"}

    f2 = LiteFeature(name="f2", primitive=Absolute(), base_features=[f1])

    column_schema = f2.column_schema
    assert column_schema.semantic_tags == {"numeric"}


def test_lite_feature_to_dependent_primitives():
    f1 = LiteFeature(name="f1", logical_type=Double)

    f2 = LiteFeature(name="f2", primitive=Absolute(), base_features=[f1])

    f3 = LiteFeature(name="f3", primitive=AddNumeric(), base_features=[f1, f2])

    f4 = LiteFeature(name="f4", primitive=MultiplyNumeric(), base_features=[f1, f3])

    assert set([x.name for x in f4.dependent_primitives()]) == set(
        ["multiply_numeric", "absolute", "add_numeric"],
    )



# flake8: noqa
from featuretools.entityset.deserialize import read_entityset
from featuretools.entityset.entityset import EntitySet
from featuretools.entityset.relationship import Relationship
from featuretools.entityset.timedelta import Timedelta



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Double

from featuretools.primitives.base import TransformPrimitive


class RateOfChange(TransformPrimitive):
    """Computes the rate of change of a value per second.

    Examples:
        >>> import pandas as pd
        >>> rate_of_change = RateOfChange()
        >>> times = pd.date_range(start='2019-01-01', freq='1min', periods=5)
        >>> results = rate_of_change([0, 30, 180, -90, 0], times).tolist()
        >>> results = [round(x, 2) for x in results]
        >>> results
        [nan, 0.5, 2.5, -4.5, 1.5]
    """

    name = "rate_of_change"
    input_types = [
        ColumnSchema(semantic_tags={"numeric"}),
        ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"}),
    ]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})
    uses_full_dataframe = True
    description_template = "the rate of change of {} per second"

    def get_function(self):
        def rate_of_change(values, time):
            time_delta = time.diff().dt.total_seconds()
            value_delta = values.diff()
            return value_delta / time_delta

        return rate_of_change



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable

from featuretools.primitives.base import TransformPrimitive


class SameAsPrevious(TransformPrimitive):
    """Determines if a value is equal to the previous value in a list.

    Description:
        Compares a value in a list to the previous value and returns True if
        the value is equal to the previous value or False otherwise. The
        first item in the output will always be False, since there is no previous
        element for the first element comparison.

        Any nan values in the input will be filled using either a forward-fill
        or backward-fill method, specified by the fill_method argument. The number
        of consecutive nan values that get filled can be limited with the limit
        argument. Any nan values left after filling will result in False being
        returned for any comparison involving the nan value.

    Args:
        fill_method (str): Method for filling gaps in series. Valid
        options are `backfill`, `bfill`, `pad`, `ffill`.
        `pad / ffill`: fill gap with last valid observation.
        `backfill / bfill`: fill gap with next valid observation.
        Default is `pad`.

        limit (int): The max number of consecutive NaN values in a gap that
            can be filled. Default is None.

    Examples:
        >>> same_as_previous = SameAsPrevious()
        >>> same_as_previous([1, 2, 2, 4]).tolist()
        [False, False, True, False]

        The fill method for nan values can be specified

        >>> same_as_previous_fillna = SameAsPrevious(fill_method="bfill")
        >>> same_as_previous_fillna([1, None, 2, 4]).tolist()
        [False, False, True, False]

        The number of nan values that are filled can be limited

        >>> same_as_previous_limitfill = SameAsPrevious(limit=2)
        >>> same_as_previous_limitfill([1, None, None, None, 2, 3]).tolist()
        [False, True, True, False, False, False]
    """

    name = "same_as_previous"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(BooleanNullable)

    def __init__(self, fill_method="pad", limit=None):
        if fill_method not in ["backfill", "bfill", "pad", "ffill"]:
            raise ValueError("Invalid fill_method")
        self.fill_method = fill_method
        self.limit = limit

    def get_function(self):
        def same_as_previous(x):
            x = x.fillna(method=self.fill_method, limit=self.limit)
            x = x.eq(x.shift())
            # first value will always be false, since there is no previous value
            x.iloc[0] = False
            return x

        return same_as_previous



import numpy as np
import pandas as pd
import pytest
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean, BooleanNullable, NaturalLanguage

from featuretools import EntitySet, Feature, dfs
from featuretools.selection import (
    remove_highly_correlated_features,
    remove_highly_null_features,
    remove_low_information_features,
    remove_single_value_features,
)
from featuretools.tests.testing_utils import make_ecommerce_entityset


@pytest.fixture
def feature_matrix():
    feature_matrix = pd.DataFrame(
        {
            "test": [0, 1, 2],
            "no_null": [np.nan, 0, 0],
            "some_null": [np.nan, 0, 0],
            "all_null": [np.nan, np.nan, np.nan],
            "many_value": [1, 2, 3],
            "dup_value": [1, 1, 2],
            "one_value": [1, 1, 1],
        },
    )
    return feature_matrix


@pytest.fixture
def test_es(pd_es, feature_matrix):
    pd_es.add_dataframe(dataframe_name="test", dataframe=feature_matrix, index="test")
    return pd_es


# remove low information features not supported in Dask
def test_remove_low_information_feature_names(feature_matrix):
    feature_matrix = remove_low_information_features(feature_matrix)
    assert feature_matrix.shape == (3, 5)
    assert "one_value" not in feature_matrix.columns
    assert "all_null" not in feature_matrix.columns


# remove low information features not supported in Dask
def test_remove_low_information_features(test_es, feature_matrix):
    features = [Feature(test_es["test"].ww[col]) for col in test_es["test"].columns]
    feature_matrix, features = remove_low_information_features(feature_matrix, features)
    assert feature_matrix.shape == (3, 5)
    assert len(features) == 5
    for f in features:
        assert f.get_name() in feature_matrix.columns
    assert "one_value" not in feature_matrix.columns
    assert "all_null" not in feature_matrix.columns


def test_remove_highly_null_features():
    nulls_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "half_nulls": [None, None, 88, 99],
            "all_nulls": [None, None, None, None],
            "quarter": ["a", "b", None, "c"],
            "vals": [True, True, False, False],
        },
    )

    es = EntitySet("data", {"nulls": (nulls_df, "id")})
    es["nulls"].ww.set_types(
        logical_types={"all_nulls": "categorical", "quarter": "categorical"},
    )
    fm, features = dfs(
        entityset=es,
        target_dataframe_name="nulls",
        trans_primitives=["is_null"],
        max_depth=1,
    )

    with pytest.raises(
        ValueError,
        match="pct_null_threshold must be a float between 0 and 1, inclusive.",
    ):
        remove_highly_null_features(fm, pct_null_threshold=1.1)

    with pytest.raises(
        ValueError,
        match="pct_null_threshold must be a float between 0 and 1, inclusive.",
    ):
        remove_highly_null_features(fm, pct_null_threshold=-0.1)

    no_thresh = remove_highly_null_features(fm)
    no_thresh_cols = set(no_thresh.columns)
    diff = set(fm.columns) - no_thresh_cols
    assert len(diff) == 1
    assert "all_nulls" not in no_thresh_cols

    half = remove_highly_null_features(fm, pct_null_threshold=0.5)
    half_cols = set(half.columns)
    diff = set(fm.columns) - half_cols
    assert len(diff) == 2
    assert "all_nulls" not in half_cols
    assert "half_nulls" not in half_cols

    no_tolerance = remove_highly_null_features(fm, pct_null_threshold=0)
    no_tolerance_cols = set(no_tolerance.columns)
    diff = set(fm.columns) - no_tolerance_cols
    assert len(diff) == 3
    assert "all_nulls" not in no_tolerance_cols
    assert "half_nulls" not in no_tolerance_cols
    assert "quarter" not in no_tolerance_cols

    (
        with_features_param,
        with_features_param_features,
    ) = remove_highly_null_features(fm, features)
    assert len(with_features_param_features) == len(no_thresh.columns)
    for i in range(len(with_features_param_features)):
        assert with_features_param_features[i].get_name() == no_thresh.columns[i]
        assert with_features_param.columns[i] == no_thresh.columns[i]


def test_remove_single_value_features():
    same_vals_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "all_numeric": [88, 88, 88, 88],
            "with_nan": [1, 1, None, 1],
            "all_nulls": [None, None, None, None],
            "all_categorical": ["a", "a", "a", "a"],
            "all_bools": [True, True, True, True],
            "diff_vals": ["hi", "bye", "bye", "hi"],
        },
    )

    es = EntitySet("data", {"single_vals": (same_vals_df, "id")})
    es["single_vals"].ww.set_types(
        logical_types={
            "all_nulls": "categorical",
            "all_categorical": "categorical",
            "diff_vals": "categorical",
        },
    )
    fm, features = dfs(
        entityset=es,
        target_dataframe_name="single_vals",
        trans_primitives=["is_null"],
        max_depth=1,
    )

    no_params, no_params_features = remove_single_value_features(fm, features)
    no_params_cols = set(no_params.columns)
    assert len(no_params_features) == 2
    assert "IS_NULL(with_nan)" in no_params_cols
    assert "diff_vals" in no_params_cols

    nan_as_value, nan_as_value_features = remove_single_value_features(
        fm,
        features,
        count_nan_as_value=True,
    )
    nan_cols = set(nan_as_value.columns)
    assert len(nan_as_value_features) == 3
    assert "IS_NULL(with_nan)" in nan_cols
    assert "diff_vals" in nan_cols
    assert "with_nan" in nan_cols

    without_features_param = remove_single_value_features(fm)
    assert len(no_params.columns) == len(without_features_param.columns)
    for i in range(len(no_params.columns)):
        assert no_params.columns[i] == without_features_param.columns[i]
        assert no_params_features[i].get_name() == without_features_param.columns[i]


def test_remove_highly_correlated_features():
    correlated_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "diff_ints": [34, 11, 29, 91],
            "words": ["test", "this is a short sentence", "foo bar", "baz"],
            "corr_words": [4, 24, 7, 3],
            "corr_1": [99, 88, 77, 33],
            "corr_2": [99, 88, 77, 33],
        },
    )

    es = EntitySet(
        "data",
        {"correlated": (correlated_df, "id", None, {"words": NaturalLanguage})},
    )
    fm, _ = dfs(
        entityset=es,
        target_dataframe_name="correlated",
        trans_primitives=["num_characters"],
        max_depth=1,
    )

    with pytest.raises(
        ValueError,
        match="pct_corr_threshold must be a float between 0 and 1, inclusive.",
    ):
        remove_highly_correlated_features(fm, pct_corr_threshold=1.1)

    with pytest.raises(
        ValueError,
        match="pct_corr_threshold must be a float between 0 and 1, inclusive.",
    ):
        remove_highly_correlated_features(fm, pct_corr_threshold=-0.1)

    with pytest.raises(
        AssertionError,
        match="feature named not_a_feature is not in feature matrix",
    ):
        remove_highly_correlated_features(fm, features_to_check=["not_a_feature"])

    to_check = remove_highly_correlated_features(
        fm,
        features_to_check=["corr_words", "NUM_CHARACTERS(words)", "diff_ints"],
    )
    to_check_columns = set(to_check.columns)
    assert len(to_check_columns) == 4
    assert "NUM_CHARACTERS(words)" not in to_check_columns
    assert "corr_1" in to_check_columns
    assert "corr_2" in to_check_columns

    to_keep = remove_highly_correlated_features(
        fm,
        features_to_keep=["NUM_CHARACTERS(words)"],
    )
    to_keep_names = set(to_keep.columns)
    assert len(to_keep_names) == 4
    assert "corr_words" in to_keep_names
    assert "NUM_CHARACTERS(words)" in to_keep_names
    assert "corr_2" not in to_keep_names

    new_fm = remove_highly_correlated_features(fm)
    assert len(new_fm.columns) == 3
    assert "corr_2" not in new_fm.columns
    assert "NUM_CHARACTERS(words)" not in new_fm.columns

    diff_threshold = remove_highly_correlated_features(fm, pct_corr_threshold=0.8)
    diff_threshold_cols = diff_threshold.columns
    assert len(diff_threshold_cols) == 2
    assert "corr_words" in diff_threshold_cols
    assert "diff_ints" in diff_threshold_cols


def test_remove_highly_correlated_features_init_woodwork():
    correlated_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "diff_ints": [34, 11, 29, 91],
            "words": ["test", "this is a short sentence", "foo bar", "baz"],
            "corr_words": [4, 24, 7, 3],
            "corr_1": [99, 88, 77, 33],
            "corr_2": [99, 88, 77, 33],
        },
    )

    es = EntitySet(
        "data",
        {"correlated": (correlated_df, "id", None, {"words": NaturalLanguage})},
    )
    fm, _ = dfs(
        entityset=es,
        target_dataframe_name="correlated",
        trans_primitives=["num_characters"],
        max_depth=1,
    )

    no_ww_fm = fm.copy()
    ww_fm = fm.copy()
    ww_fm.ww.init()

    new_no_ww_fm = remove_highly_correlated_features(no_ww_fm)
    new_ww_fm = remove_highly_correlated_features(ww_fm)

    pd.testing.assert_frame_equal(new_no_ww_fm, new_ww_fm)


def test_multi_output_selection():
    df1 = pd.DataFrame({"id": [0, 1, 2, 3]})

    df2 = pd.DataFrame(
        {
            "index": [0, 1, 2, 3],
            "first_id": [0, 1, 1, 3],
            "all_nulls": [None, None, None, None],
            "quarter": ["a", "b", None, "c"],
        },
    )

    dataframes = {
        "first": (df1, "id"),
        "second": (df2, "index"),
    }

    relationships = [("first", "id", "second", "first_id")]
    es = EntitySet("data", dataframes, relationships=relationships)
    es["second"].ww.set_types(
        logical_types={"all_nulls": "categorical", "quarter": "categorical"},
    )

    fm, features = dfs(
        entityset=es,
        target_dataframe_name="first",
        trans_primitives=[],
        agg_primitives=["n_most_common"],
        max_depth=1,
    )

    multi_output, multi_output_features = remove_single_value_features(fm, features)
    assert multi_output.columns == ["N_MOST_COMMON(second.quarter)[0]"]
    assert len(multi_output_features) == 1
    assert multi_output_features[0].get_name() == multi_output.columns[0]

    es = make_ecommerce_entityset()
    fm, features = dfs(
        entityset=es,
        target_dataframe_name="régions",
        trans_primitives=[],
        agg_primitives=["n_most_common"],
        max_depth=2,
    )

    matrix_with_slices, unsliced_features = remove_highly_null_features(fm, features)
    assert len(matrix_with_slices.columns) == 18
    assert len(unsliced_features) == 14

    matrix_columns = set(matrix_with_slices.columns)
    for f in unsliced_features:
        for f_name in f.get_feature_names():
            assert f_name in matrix_columns


def test_remove_highly_correlated_features_on_boolean_cols():
    correlated_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "diff_ints": [34, 11, 29, 91],
            "corr_words": [4, 24, 7, 3],
            "bools": [True, True, False, True],
        },
    )

    es = EntitySet(
        "data",
        {"correlated": (correlated_df, "id", None, {"bools": Boolean})},
    )

    feature_matrix, features = dfs(
        entityset=es,
        target_dataframe_name="correlated",
        trans_primitives=["equal"],
        agg_primitives=[],
        max_depth=1,
        return_types=[
            ColumnSchema(logical_type=BooleanNullable),
            ColumnSchema(logical_type=Boolean),
        ],
    )
    # Confirm both boolean logical types are included so that we know we're checking the correct types
    assert {
        ltype.type_string for ltype in feature_matrix.ww.logical_types.values()
    } == {Boolean.type_string, BooleanNullable.type_string}

    to_keep = remove_highly_correlated_features(
        feature_matrix=feature_matrix,
        features=features,
        pct_corr_threshold=0.3,
    )
    assert len(to_keep[0].columns) < len(feature_matrix.columns)



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class GreaterThanScalar(TransformPrimitive):
    """Determines if values are greater than a given scalar.

    Description:
        Given a list of values and a constant scalar, determine
        whether each of the values is greater than the scalar.
        If a value is equal to the scalar, return `False`.

    Examples:
        >>> greater_than_scalar = GreaterThanScalar(value=2)
        >>> greater_than_scalar([3, 1, 2]).tolist()
        [True, False, False]
    """

    name = "greater_than_scalar"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]

    def __init__(self, value=0):
        self.value = value
        self.description_template = "whether {{}} is greater than {}".format(self.value)

    def get_function(self):
        def greater_than_scalar(vals):
            return vals > self.value

        return greater_than_scalar

    def generate_name(self, base_feature_names):
        return "%s > %s" % (base_feature_names[0], str(self.value))



from featuretools.primitives.standard.transform.numeric.absolute import Absolute
from featuretools.primitives.standard.transform.numeric.cosine import Cosine
from featuretools.primitives.standard.transform.numeric.diff import Diff
from featuretools.primitives.standard.transform.numeric.natural_logarithm import (
    NaturalLogarithm,
)
from featuretools.primitives.standard.transform.numeric.negate import Negate
from featuretools.primitives.standard.transform.numeric.percentile import Percentile
from featuretools.primitives.standard.transform.numeric.rate_of_change import (
    RateOfChange,
)
from featuretools.primitives.standard.transform.numeric.same_as_previous import (
    SameAsPrevious,
)
from featuretools.primitives.standard.transform.numeric.sine import Sine
from featuretools.primitives.standard.transform.numeric.square_root import SquareRoot
from featuretools.primitives.standard.transform.numeric.tangent import Tangent



from featuretools.primitives.standard.transform.postal.one_digit_postal_code import (
    OneDigitPostalCode,
)
from featuretools.primitives.standard.transform.postal.two_digit_postal_code import (
    TwoDigitPostalCode,
)



from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from featuretools.primitives import (
    RollingCount,
    RollingMax,
    RollingMean,
    RollingMin,
    RollingSTD,
    RollingTrend,
)
from featuretools.primitives.standard.transform.time_series.utils import (
    _get_rolled_series_without_gap,
    apply_roll_with_offset_gap,
    roll_series_with_gap,
)
from featuretools.tests.primitive_tests.utils import get_number_from_offset


def test_get_rolled_series_without_gap(window_series_pd):
    # Data is daily, so number of rows should be number of days not included in the gap
    assert len(_get_rolled_series_without_gap(window_series_pd, "11D")) == 9
    assert len(_get_rolled_series_without_gap(window_series_pd, "0D")) == 20
    assert len(_get_rolled_series_without_gap(window_series_pd, "48H")) == 18
    assert len(_get_rolled_series_without_gap(window_series_pd, "4H")) == 19


def test_get_rolled_series_without_gap_not_uniform(window_series_pd):
    non_uniform_series = window_series_pd.iloc[[0, 2, 5, 6, 8, 9]]

    assert len(_get_rolled_series_without_gap(non_uniform_series, "10D")) == 0
    assert len(_get_rolled_series_without_gap(non_uniform_series, "0D")) == 6
    assert len(_get_rolled_series_without_gap(non_uniform_series, "48H")) == 4
    assert len(_get_rolled_series_without_gap(non_uniform_series, "4H")) == 5
    assert len(_get_rolled_series_without_gap(non_uniform_series, "4D")) == 3
    assert len(_get_rolled_series_without_gap(non_uniform_series, "4D2H")) == 2


def test_get_rolled_series_without_gap_empty_series(window_series_pd):
    empty_series = pd.Series([], dtype="object")
    assert len(_get_rolled_series_without_gap(empty_series, "1D")) == 0
    assert len(_get_rolled_series_without_gap(empty_series, "0D")) == 0


def test_get_rolled_series_without_gap_large_bound(window_series_pd):
    assert len(_get_rolled_series_without_gap(window_series_pd, "100D")) == 0
    assert (
        len(
            _get_rolled_series_without_gap(
                window_series_pd.iloc[[0, 2, 5, 6, 8, 9]],
                "20D",
            ),
        )
        == 0
    )


@pytest.mark.parametrize(
    "window_length, gap",
    [
        (3, 2),
        (3, 4),  # gap larger than window
        (2, 0),  # gap explicitly set to 0
        ("3d", "2d"),  # using offset aliases
        ("3d", "4d"),
        ("4d", "0d"),
    ],
)
def test_roll_series_with_gap(window_length, gap, window_series_pd):
    rolling_max = roll_series_with_gap(
        window_series_pd,
        window_length,
        gap=gap,
        min_periods=1,
    ).max()
    rolling_min = roll_series_with_gap(
        window_series_pd,
        window_length,
        gap=gap,
        min_periods=1,
    ).min()

    assert len(rolling_max) == len(window_series_pd)
    assert len(rolling_min) == len(window_series_pd)

    gap_num = get_number_from_offset(gap)
    window_length_num = get_number_from_offset(window_length)
    for i in range(len(window_series_pd)):
        start_idx = i - gap_num - window_length_num + 1

        if isinstance(gap, str):
            # No gap functionality is happening, so gap isn't taken account in the end index
            # it's like the gap is 0; it includes the row itself
            end_idx = i
        else:
            end_idx = i - gap_num

        # If start and end are negative, they're entirely before
        if start_idx < 0 and end_idx < 0:
            assert pd.isnull(rolling_max.iloc[i])
            assert pd.isnull(rolling_min.iloc[i])
            continue

        if start_idx < 0:
            start_idx = 0

        # Because the row values are a range from 0 to 20, the rolling min will be the start index
        # and the rolling max will be the end idx
        assert rolling_min.iloc[i] == start_idx
        assert rolling_max.iloc[i] == end_idx


@pytest.mark.parametrize("window_length", [3, "3d"])
def test_roll_series_with_no_gap(window_length, window_series_pd):
    actual_rolling = roll_series_with_gap(
        window_series_pd,
        window_length,
        gap=0,
        min_periods=1,
    ).mean()
    expected_rolling = window_series_pd.rolling(window_length, min_periods=1).mean()

    pd.testing.assert_series_equal(actual_rolling, expected_rolling)


@pytest.mark.parametrize(
    "window_length, gap",
    [
        (6, 2),
        (6, 0),  # No gap - changes early values
        ("6d", "0d"),  # Uses offset aliases
        ("6d", "2d"),
    ],
)
def test_roll_series_with_gap_early_values(window_length, gap, window_series_pd):
    gap_num = get_number_from_offset(gap)
    window_length_num = get_number_from_offset(window_length)

    # Default min periods is 1 - will include all
    default_partial_values = roll_series_with_gap(
        window_series_pd,
        window_length,
        gap=gap,
        min_periods=1,
    ).count()
    num_empty_aggregates = len(default_partial_values.loc[default_partial_values == 0])
    num_partial_aggregates = len(
        (default_partial_values.loc[default_partial_values != 0]).loc[
            default_partial_values < window_length_num
        ],
    )

    assert num_partial_aggregates == window_length_num - 1
    if isinstance(gap, str):
        # gap isn't handled, so we'll always at least include the row itself
        assert num_empty_aggregates == 0
    else:
        assert num_empty_aggregates == gap_num

    # Make min periods the size of the window
    no_partial_values = roll_series_with_gap(
        window_series_pd,
        window_length,
        gap=gap,
        min_periods=window_length_num,
    ).count()
    num_null_aggregates = len(no_partial_values.loc[pd.isna(no_partial_values)])
    num_partial_aggregates = len(
        no_partial_values.loc[no_partial_values < window_length_num],
    )

    # because we shift, gap is included as nan values in the series.
    # Count treats nans in a window as values that don't get counted,
    # so the gap rows get included in the count for whether a window has "min periods".
    # This is different than max, for example, which does not count nans in a window as values towards "min periods"
    assert num_null_aggregates == window_length_num - 1
    if isinstance(gap, str):
        # gap isn't handled, so we'll never have any partial aggregates
        assert num_partial_aggregates == 0
    else:
        assert num_partial_aggregates == gap_num


def test_roll_series_with_gap_nullable_types(window_series_pd):
    window_length = 3
    gap = 2
    min_periods = 1
    # Because we're inserting nans, confirm that nullability of the dtype doesn't have an impact on the results
    nullable_series = window_series_pd.astype("Int64")
    non_nullable_series = window_series_pd.astype("int64")

    nullable_rolling_max = roll_series_with_gap(
        nullable_series,
        window_length,
        gap=gap,
        min_periods=min_periods,
    ).max()
    non_nullable_rolling_max = roll_series_with_gap(
        non_nullable_series,
        window_length,
        gap=gap,
        min_periods=min_periods,
    ).max()

    pd.testing.assert_series_equal(nullable_rolling_max, non_nullable_rolling_max)


def test_roll_series_with_gap_nullable_types_with_nans(window_series_pd):
    window_length = 3
    gap = 2
    min_periods = 1
    nullable_floats = window_series_pd.astype("float64").replace(
        {1: np.nan, 3: np.nan},
    )
    nullable_ints = nullable_floats.astype("Int64")

    nullable_ints_rolling_max = roll_series_with_gap(
        nullable_ints,
        window_length,
        gap=gap,
        min_periods=min_periods,
    ).max()
    nullable_floats_rolling_max = roll_series_with_gap(
        nullable_floats,
        window_length,
        gap=gap,
        min_periods=min_periods,
    ).max()

    pd.testing.assert_series_equal(
        nullable_ints_rolling_max,
        nullable_floats_rolling_max,
    )

    expected_early_values = [np.nan, np.nan, 0, 0, 2, 2, 4] + list(
        range(7 - gap, len(window_series_pd) - gap),
    )
    for i in range(len(window_series_pd)):
        actual = nullable_floats_rolling_max.iloc[i]
        expected = expected_early_values[i]

        if pd.isnull(actual):
            assert pd.isnull(expected)
        else:
            assert actual == expected


@pytest.mark.parametrize(
    "window_length, gap",
    [
        ("3d", "2d"),
        ("3d", "4d"),
        ("4d", "0d"),
    ],
)
def test_apply_roll_with_offset_gap(window_length, gap, window_series_pd):
    def max_wrapper(sub_s):
        return apply_roll_with_offset_gap(sub_s, gap, max, min_periods=1)

    rolling_max_obj = roll_series_with_gap(
        window_series_pd,
        window_length,
        gap=gap,
        min_periods=1,
    )
    rolling_max_series = rolling_max_obj.apply(max_wrapper)

    def min_wrapper(sub_s):
        return apply_roll_with_offset_gap(sub_s, gap, min, min_periods=1)

    rolling_min_obj = roll_series_with_gap(
        window_series_pd,
        window_length,
        gap=gap,
        min_periods=1,
    )
    rolling_min_series = rolling_min_obj.apply(min_wrapper)

    assert len(rolling_max_series) == len(window_series_pd)
    assert len(rolling_min_series) == len(window_series_pd)

    gap_num = get_number_from_offset(gap)
    window_length_num = get_number_from_offset(window_length)
    for i in range(len(window_series_pd)):
        start_idx = i - gap_num - window_length_num + 1
        # Now that we have the _apply call, this acts as expected
        end_idx = i - gap_num

        # If start and end are negative, they're entirely before
        if start_idx < 0 and end_idx < 0:
            assert pd.isnull(rolling_max_series.iloc[i])
            assert pd.isnull(rolling_min_series.iloc[i])
            continue

        if start_idx < 0:
            start_idx = 0

        # Because the row values are a range from 0 to 20, the rolling min will be the start index
        # and the rolling max will be the end idx
        assert rolling_min_series.iloc[i] == start_idx
        assert rolling_max_series.iloc[i] == end_idx


@pytest.mark.parametrize(
    "min_periods",
    [1, 0, None],
)
def test_apply_roll_with_offset_gap_default_min_periods(min_periods, window_series_pd):
    window_length = "5d"
    window_length_num = 5
    gap = "3d"
    gap_num = 3

    def count_wrapper(sub_s):
        return apply_roll_with_offset_gap(sub_s, gap, len, min_periods=min_periods)

    rolling_count_obj = roll_series_with_gap(
        window_series_pd,
        window_length,
        gap=gap,
        min_periods=min_periods,
    )
    rolling_count_series = rolling_count_obj.apply(count_wrapper)

    # gap essentially creates a rolling series that has no elements; which should be nan
    # to differentiate from when a window only has null values
    num_empty_aggregates = rolling_count_series.isna().sum()
    num_partial_aggregates = len(
        (rolling_count_series.loc[rolling_count_series != 0]).loc[
            rolling_count_series < window_length_num
        ],
    )

    assert num_empty_aggregates == gap_num
    assert num_partial_aggregates == window_length_num - 1


@pytest.mark.parametrize(
    "min_periods",
    [2, 3, 4, 5],
)
def test_apply_roll_with_offset_gap_min_periods(min_periods, window_series_pd):
    window_length = "5d"
    window_length_num = 5
    gap = "3d"
    gap_num = 3

    def count_wrapper(sub_s):
        return apply_roll_with_offset_gap(sub_s, gap, len, min_periods=min_periods)

    rolling_count_obj = roll_series_with_gap(
        window_series_pd,
        window_length,
        gap=gap,
        min_periods=min_periods,
    )
    rolling_count_series = rolling_count_obj.apply(count_wrapper)

    # gap essentially creates rolling series that have no elements; which should be nan
    # to differentiate from when a window only has null values
    num_empty_aggregates = rolling_count_series.isna().sum()
    num_partial_aggregates = len(
        (rolling_count_series.loc[rolling_count_series != 0]).loc[
            rolling_count_series < window_length_num
        ],
    )

    assert num_empty_aggregates == min_periods - 1 + gap_num
    assert num_partial_aggregates == window_length_num - min_periods


def test_apply_roll_with_offset_gap_non_uniform():
    window_length = "3d"
    gap = "3d"
    min_periods = 1
    # When the data isn't uniform, this impacts the number of values in each rolling window
    datetimes = (
        list(pd.date_range(start="2017-01-01", freq="1d", periods=7))
        + list(pd.date_range(start="2017-02-01", freq="2d", periods=7))
        + list(pd.date_range(start="2017-03-01", freq="1d", periods=7))
    )
    no_freq_series = pd.Series(range(len(datetimes)), index=datetimes)

    assert pd.infer_freq(no_freq_series.index) is None

    expected_series = pd.Series(
        [None, None, None, 1, 2, 3, 3]
        + [None, None, 1, 1, 1, 1, 1]
        + [None, None, None, 1, 2, 3, 3],
        index=datetimes,
    )

    def count_wrapper(sub_s):
        return apply_roll_with_offset_gap(sub_s, gap, len, min_periods=min_periods)

    rolling_count_obj = roll_series_with_gap(
        no_freq_series,
        window_length,
        gap=gap,
        min_periods=min_periods,
    )
    rolling_count_series = rolling_count_obj.apply(count_wrapper)

    pd.testing.assert_series_equal(rolling_count_series, expected_series)


def test_apply_roll_with_offset_data_frequency_higher_than_parameters_frequency():
    window_length = "5D"  # 120 hours
    window_length_num = 5
    # In order for min periods to be the length of the window, we multiply 24hours*5
    min_periods = window_length_num * 24

    datetimes = list(pd.date_range(start="2017-01-01", freq="1H", periods=200))
    high_frequency_series = pd.Series(range(200), index=datetimes)

    # Check without gap
    gap = "0d"
    gap_num = 0

    def max_wrapper(sub_s):
        return apply_roll_with_offset_gap(sub_s, gap, max, min_periods=min_periods)

    rolling_max_obj = roll_series_with_gap(
        high_frequency_series,
        window_length,
        min_periods=min_periods,
        gap=gap,
    )
    rolling_max_series = rolling_max_obj.apply(max_wrapper)

    assert rolling_max_series.isna().sum() == (min_periods - 1) + gap_num

    # Check with small gap
    gap = "3H"
    gap_num = 3

    def max_wrapper(sub_s):
        return apply_roll_with_offset_gap(sub_s, gap, max, min_periods=min_periods)

    rolling_max_obj = roll_series_with_gap(
        high_frequency_series,
        window_length,
        min_periods=min_periods,
        gap=gap,
    )
    rolling_max_series = rolling_max_obj.apply(max_wrapper)

    assert rolling_max_series.isna().sum() == (min_periods - 1) + gap_num

    # Check with large gap - in terms of days, so we'll multiply by 24hours for number of nans
    gap = "2D"
    gap_num = 2

    def max_wrapper(sub_s):
        return apply_roll_with_offset_gap(sub_s, gap, max, min_periods=min_periods)

    rolling_max_obj = roll_series_with_gap(
        high_frequency_series,
        window_length,
        min_periods=min_periods,
        gap=gap,
    )
    rolling_max_series = rolling_max_obj.apply(max_wrapper)

    assert rolling_max_series.isna().sum() == (min_periods - 1) + (gap_num * 24)


def test_apply_roll_with_offset_data_min_periods_too_big(window_series_pd):
    window_length = "5D"
    gap = "2d"

    # Since the data has a daily frequency, there will only be, at most, 5 rows in the window
    min_periods = 6

    def max_wrapper(sub_s):
        return apply_roll_with_offset_gap(sub_s, gap, max, min_periods=min_periods)

    rolling_max_obj = roll_series_with_gap(
        window_series_pd,
        window_length,
        min_periods=min_periods,
        gap=gap,
    )
    rolling_max_series = rolling_max_obj.apply(max_wrapper)

    # The resulting series is comprised entirely of nans
    assert rolling_max_series.isna().sum() == len(window_series_pd)


def test_roll_series_with_gap_different_input_types_same_result_uniform(
    window_series_pd,
):
    # Offset inputs will only produce the same results as numeric inputs
    # when the data has a uniform frequency
    offset_gap = "2d"
    offset_window_length = "5d"
    int_gap = 2
    int_window_length = 5
    min_periods = 1

    # Rolling series' with matching input types
    expected_rolling_numeric = roll_series_with_gap(
        window_series_pd,
        window_length=int_window_length,
        gap=int_gap,
        min_periods=min_periods,
    ).max()

    def count_wrapper(sub_s):
        return apply_roll_with_offset_gap(
            sub_s,
            offset_gap,
            max,
            min_periods=min_periods,
        )

    rolling_count_obj = roll_series_with_gap(
        window_series_pd,
        window_length=offset_window_length,
        gap=offset_gap,
        min_periods=min_periods,
    )
    expected_rolling_offset = rolling_count_obj.apply(count_wrapper)

    # confirm that the offset and gap results are equal to one another
    pd.testing.assert_series_equal(expected_rolling_numeric, expected_rolling_offset)

    # Rolling series' with mismatched input types
    mismatched_numeric_gap = roll_series_with_gap(
        window_series_pd,
        window_length=offset_window_length,
        gap=int_gap,
        min_periods=min_periods,
    ).max()
    # Confirm the mismatched results also produce the same results
    pd.testing.assert_series_equal(expected_rolling_numeric, mismatched_numeric_gap)


def test_roll_series_with_gap_incorrect_types(window_series_pd):
    error = "Window length must be either an offset string or an integer."
    with pytest.raises(TypeError, match=error):
        roll_series_with_gap(
            window_series_pd,
            window_length=4.2,
            gap=4,
            min_periods=1,
        ),

    error = "Gap must be either an offset string or an integer."
    with pytest.raises(TypeError, match=error):
        roll_series_with_gap(window_series_pd, window_length=4, gap=4.2, min_periods=1)


def test_roll_series_with_gap_negative_inputs(window_series_pd):
    error = "Window length must be greater than zero."
    with pytest.raises(ValueError, match=error):
        roll_series_with_gap(window_series_pd, window_length=-4, gap=4, min_periods=1)

    error = "Gap must be greater than or equal to zero."
    with pytest.raises(ValueError, match=error):
        roll_series_with_gap(window_series_pd, window_length=4, gap=-4, min_periods=1)


def test_roll_series_with_non_offset_string_inputs(window_series_pd):
    error = "Cannot roll series. The specified gap, test, is not a valid offset alias."
    with pytest.raises(ValueError, match=error):
        roll_series_with_gap(
            window_series_pd,
            window_length="4D",
            gap="test",
            min_periods=1,
        )

    error = "Cannot roll series. The specified window length, test, is not a valid offset alias."
    with pytest.raises(ValueError, match=error):
        roll_series_with_gap(
            window_series_pd,
            window_length="test",
            gap="7D",
            min_periods=1,
        )

    # Test mismatched types error
    error = (
        "Cannot roll series with offset gap, 2d, and numeric window length, 7. "
        "If an offset alias is used for gap, the window length must also be defined as an offset alias. "
        "Please either change gap to be numeric or change window length to be an offset alias."
    )
    with pytest.raises(TypeError, match=error):
        roll_series_with_gap(
            window_series_pd,
            window_length=7,
            gap="2d",
            min_periods=1,
        ).max()


@pytest.mark.parametrize(
    "primitive",
    [RollingCount, RollingMax, RollingMin, RollingMean, RollingSTD, RollingTrend],
)
@patch(
    "featuretools.primitives.standard.transform.time_series.utils.apply_roll_with_offset_gap",
)
def test_no_call_to_apply_roll_with_offset_gap_with_numeric(
    mock_apply_roll,
    primitive,
    window_series_pd,
):
    assert not mock_apply_roll.called

    fully_numeric_primitive = primitive(window_length=3, gap=1)
    primitive_func = fully_numeric_primitive.get_function()
    if isinstance(fully_numeric_primitive, RollingCount):
        pd.Series(primitive_func(window_series_pd.index))
    else:
        pd.Series(
            primitive_func(
                window_series_pd.index,
                pd.Series(window_series_pd.values),
            ),
        )

    assert not mock_apply_roll.called

    offset_window_primitive = primitive(window_length="3d", gap=1)
    primitive_func = offset_window_primitive.get_function()
    if isinstance(offset_window_primitive, RollingCount):
        pd.Series(primitive_func(window_series_pd.index))
    else:
        pd.Series(
            primitive_func(
                window_series_pd.index,
                pd.Series(window_series_pd.values),
            ),
        )

    assert not mock_apply_roll.called

    no_gap_specified_primitive = primitive(window_length="3d")
    primitive_func = no_gap_specified_primitive.get_function()
    if isinstance(no_gap_specified_primitive, RollingCount):
        pd.Series(primitive_func(window_series_pd.index))
    else:
        pd.Series(
            primitive_func(
                window_series_pd.index,
                pd.Series(window_series_pd.values),
            ),
        )

    assert not mock_apply_roll.called

    no_gap_specified_primitive = primitive(window_length="3d", gap="1d")
    primitive_func = no_gap_specified_primitive.get_function()
    if isinstance(no_gap_specified_primitive, RollingCount):
        pd.Series(primitive_func(window_series_pd.index))
    else:
        pd.Series(
            primitive_func(
                window_series_pd.index,
                pd.Series(window_series_pd.values),
            ),
        )

    assert mock_apply_roll.called



from scipy import stats
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils.gen_utils import Library


class Entropy(AggregationPrimitive):
    """Calculates the entropy for a categorical column

    Description:
        Given a list of observations from a categorical
        column return the entropy of the distribution.
        NaN values can be treated as a category or
        dropped.

    Args:
        dropna (bool): Whether to consider NaN values as a separate category
            Defaults to False.
        base (float): The logarithmic base to use
            Defaults to e (natural logarithm)

    Examples:
        >>> pd_entropy = Entropy()
        >>> pd_entropy([1, 2, 3, 4])
        1.3862943611198906
    """

    name = "entropy"
    input_types = [ColumnSchema(semantic_tags={"category"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    stack_on_self = False
    description_template = "the entropy of {}"

    def __init__(self, dropna=False, base=None):
        self.dropna = dropna
        self.base = base

    def get_function(self, agg_type=Library.PANDAS):
        def pd_entropy(s):
            distribution = s.value_counts(normalize=True, dropna=self.dropna)
            if distribution.dtype == "Float64":
                distribution = distribution.astype("float64")
            return stats.entropy(distribution.to_numpy(), base=self.base)

        return pd_entropy



import json
import os
import tarfile
import tempfile

import pandas as pd
import woodwork.type_sys.type_system as ww_type_system
from woodwork.deserialize import read_woodwork_table

from featuretools.entityset.relationship import Relationship
from featuretools.utils.gen_utils import Library, import_or_none
from featuretools.utils.s3_utils import get_transport_params, use_smartopen_es
from featuretools.utils.schema_utils import check_schema_version
from featuretools.utils.wrangle import _is_local_tar, _is_s3, _is_url

dd = import_or_none("dask.dataframe")
ps = import_or_none("pyspark.pandas")


def description_to_entityset(description, **kwargs):
    """Deserialize entityset from data description.

    Args:
        description (dict) : Description of an :class:`.EntitySet`. Likely generated using :meth:`.serialize.entityset_to_description`
        kwargs (keywords): Additional keyword arguments to pass as keywords arguments to the underlying deserialization method.

    Returns:
        entityset (EntitySet) : Instance of :class:`.EntitySet`.
    """
    check_schema_version(description, "entityset")

    from featuretools.entityset import EntitySet

    # If data description was not read from disk, path is None.
    path = description.get("path")
    entityset = EntitySet(description["id"])

    for df in description["dataframes"].values():
        if path is not None:
            data_path = os.path.join(path, "data", df["name"])
            format = description.get("format")
            if format is not None:
                kwargs["format"] = format
                if format == "parquet" and df["loading_info"]["table_type"] == "pandas":
                    kwargs["filename"] = df["name"] + ".parquet"
            dataframe = read_woodwork_table(data_path, validate=False, **kwargs)
        else:
            dataframe = empty_dataframe(df, description["data_type"])

        entityset.add_dataframe(dataframe)

    for relationship in description["relationships"]:
        rel = Relationship.from_dictionary(relationship, entityset)
        entityset.add_relationship(relationship=rel)

    return entityset


def empty_dataframe(description, data_type=Library.PANDAS):
    """Deserialize empty dataframe from dataframe description.

    Args:
        description (dict) : Description of dataframe.

    Returns:
        df (DataFrame) : Empty dataframe with Woodwork initialized.
    """
    # TODO: Can we update Woodwork to return an empty initialized dataframe from a description
    # instead of using this function? Or otherwise eliminate? Issue #1476
    logical_types = {}
    semantic_tags = {}
    column_descriptions = {}
    column_metadata = {}
    use_standard_tags = {}
    category_dtypes = {}
    columns = []
    for col in description["column_typing_info"]:
        col_name = col["name"]
        columns.append(col_name)

        ltype_metadata = col["logical_type"]
        ltype = ww_type_system.str_to_logical_type(
            ltype_metadata["type"],
            params=ltype_metadata["parameters"],
        )

        tags = col["semantic_tags"]

        if "index" in tags:
            tags.remove("index")
        elif "time_index" in tags:
            tags.remove("time_index")

        logical_types[col_name] = ltype
        semantic_tags[col_name] = tags
        column_descriptions[col_name] = col["description"]
        column_metadata[col_name] = col["metadata"]
        use_standard_tags[col_name] = col["use_standard_tags"]

        if col["physical_type"]["type"] == "category":
            # Make sure categories are recreated properly
            cat_values = col["physical_type"]["cat_values"]
            cat_dtype = col["physical_type"]["cat_dtype"]
            cat_object = pd.CategoricalDtype(pd.Index(cat_values, dtype=cat_dtype))
            category_dtypes[col_name] = cat_object

    dataframe = pd.DataFrame(columns=columns).astype(category_dtypes)
    if data_type == Library.DASK:
        dataframe = dd.from_pandas(dataframe, npartitions=1)
    elif data_type == Library.SPARK:
        dataframe = ps.from_pandas(dataframe)

    dataframe.ww.init(
        name=description.get("name"),
        index=description.get("index"),
        time_index=description.get("time_index"),
        logical_types=logical_types,
        semantic_tags=semantic_tags,
        use_standard_tags=use_standard_tags,
        table_metadata=description.get("table_metadata"),
        column_metadata=column_metadata,
        column_descriptions=column_descriptions,
        validate=False,
    )

    return dataframe


def read_data_description(path):
    """Read data description from disk, S3 path, or URL.

    Args:
        path (str): Location on disk, S3 path, or URL to read `data_description.json`.

    Returns:
        description (dict) : Description of :class:`.EntitySet`.
    """

    path = os.path.abspath(path)
    assert os.path.exists(path), '"{}" does not exist'.format(path)
    filepath = os.path.join(path, "data_description.json")
    with open(filepath, "r") as file:
        description = json.load(file)
    description["path"] = path
    return description


def read_entityset(path, profile_name=None, **kwargs):
    """Read entityset from disk, S3 path, or URL.

    Args:
        path (str): Directory on disk, S3 path, or URL to read `data_description.json`.
        profile_name (str, bool): The AWS profile specified to write to S3. Will default to None and search for AWS credentials.
            Set to False to use an anonymous profile.
        kwargs (keywords): Additional keyword arguments to pass as keyword arguments to the underlying deserialization method.
    """
    if _is_url(path) or _is_s3(path) or _is_local_tar(str(path)):
        with tempfile.TemporaryDirectory() as tmpdir:
            local_path = path
            transport_params = None

            if _is_s3(path):
                transport_params = get_transport_params(profile_name)

            if _is_s3(path) or _is_url(path):
                local_path = os.path.join(tmpdir, "temporary_es")
                use_smartopen_es(local_path, path, transport_params)

            with tarfile.open(str(local_path)) as tar:
                tar.extractall(path=tmpdir)

            data_description = read_data_description(tmpdir)
            return description_to_entityset(data_description, **kwargs)
    else:
        data_description = read_data_description(path)
        return description_to_entityset(data_description, **kwargs)



import numpy as np
import pandas as pd
import pytest

from featuretools.primitives import MeanCharactersPerWord
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestMeanCharactersPerWord(PrimitiveTestBase):
    primitive = MeanCharactersPerWord

    def test_sentences(self):
        x = pd.Series(
            [
                "This is a test file",
                "This is second line",
                "third line $1,000",
                "and subsequent lines",
                "and more",
            ],
        )
        primitive_func = self.primitive().get_function()
        answers = pd.Series([3.0, 4.0, 5.0, 6.0, 3.5])
        pd.testing.assert_series_equal(primitive_func(x), answers, check_names=False)

    def test_punctuation(self):
        x = pd.Series(
            [
                "This: is a test file",
                "This, is second line?",
                "third/line $1,000;",
                "and--subsequen't lines...",
                "*and, more..",
            ],
        )
        primitive_func = self.primitive().get_function()
        answers = pd.Series([3.0, 4.0, 8.0, 10.5, 4.0])
        pd.testing.assert_series_equal(primitive_func(x), answers, check_names=False)

    def test_multiline(self):
        x = pd.Series(
            [
                "This is a test file",
                "This is second line\nthird line $1000;\nand subsequent lines",
                "and more",
            ],
        )
        primitive_func = self.primitive().get_function()
        answers = pd.Series([3.0, 4.8, 3.5])
        pd.testing.assert_series_equal(primitive_func(x), answers, check_names=False)

    @pytest.mark.parametrize(
        "na_value",
        [None, np.nan, pd.NA],
    )
    def test_nans(self, na_value):
        x = pd.Series([na_value, "", "third line"])
        primitive_func = self.primitive().get_function()
        answers = pd.Series([np.nan, 0, 4.5])
        pd.testing.assert_series_equal(primitive_func(x), answers, check_names=False)

    @pytest.mark.parametrize(
        "na_value",
        [None, np.nan, pd.NA],
    )
    def test_all_nans(self, na_value):
        x = pd.Series([na_value, na_value, na_value])
        primitive_func = self.primitive().get_function()
        answers = pd.Series([np.nan, np.nan, np.nan])
        pd.testing.assert_series_equal(primitive_func(x), answers, check_names=False)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import URL, Categorical

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.common_tld_utils import COMMON_TLDS


class URLToTLD(TransformPrimitive):
    """Determines the top level domain of a url.

    Description:
        Extract the top level domain of a url, using regex,
        and a list of common top level domains. Returns nan if
        the url is invalid or null.
        Common top level domains were pulled from this list:
        https://www.hayksaakian.com/most-popular-tlds/

    Examples:
        >>> url_to_tld = URLToTLD()
        >>> urls = ['https://www.google.com', 'http://www.google.co.in',
        ...         'www.facebook.com']
        >>> url_to_tld(urls).to_list()
        ['com', 'in', 'com']
    """

    name = "url_to_tld"
    input_types = [ColumnSchema(logical_type=URL)]
    return_type = ColumnSchema(logical_type=Categorical, semantic_tags={"category"})

    def get_function(self):
        self.tlds_pattern = r"(?:\.({}))".format("|".join(COMMON_TLDS))

        def url_to_domain(x):
            p = r"^(?:https?:\/\/)?(?:[^@\/\n]+@)?(?:www\.)?([^:\/?\n]+)"
            return x.str.extract(p, expand=False)

        def url_to_tld(x):
            domains = url_to_domain(x)
            df = domains.str.extractall(self.tlds_pattern)
            matches = df.groupby(level=0).last()[0]
            return matches.reindex(x.index)

        return url_to_tld



import numpy as np
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base import TransformPrimitive


class CumMean(TransformPrimitive):
    """Calculates the cumulative mean.

    Description:
        Given a list of values, return the cumulative mean
        (or running mean). There is no set window, so the
        mean at each point is calculated over all prior values.
        `NaN` values will return `NaN`, but in the window of a
        cumulative caluclation, they're treated as 0.

    Examples:
        >>> cum_mean = CumMean()
        >>> cum_mean([1, 2, 3, 4, None, 5]).tolist()
        [1.0, 1.5, 2.0, 2.5, nan, 2.5]
    """

    name = "cum_mean"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    uses_full_dataframe = True
    description_template = "the cumulative mean of {}"

    def get_function(self):
        def cum_mean(values):
            return values.cumsum() / np.arange(1, len(values) + 1)

        return cum_mean



import numpy as np
import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime

from featuretools import (
    Feature,
    GroupByTransformFeature,
    IdentityFeature,
    calculate_feature_matrix,
    feature_base,
)
from featuretools.computational_backends.feature_set import FeatureSet
from featuretools.computational_backends.feature_set_calculator import (
    FeatureSetCalculator,
)
from featuretools.primitives import CumCount, CumMax, CumMean, CumMin, CumSum, Last
from featuretools.primitives.base import TransformPrimitive
from featuretools.synthesis import dfs
from featuretools.tests.testing_utils import feature_with_name


class TestCumCount:
    primitive = CumCount

    def test_order(self):
        g = pd.Series(["a", "b", "a"])

        answers = ([1, 2], [1])

        function = self.primitive().get_function()
        for (_, group), answer in zip(g.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)

    def test_regular(self):
        g = pd.Series(["a", "b", "a", "c", "d", "b"])
        answers = ([1, 2], [1, 2], [1], [1])

        function = self.primitive().get_function()
        for (_, group), answer in zip(g.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)

    def test_discrete(self):
        g = pd.Series(["a", "b", "a", "c", "d", "b"])
        answers = ([1, 2], [1, 2], [1], [1])

        function = self.primitive().get_function()
        for (_, group), answer in zip(g.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)


class TestCumSum:
    primitive = CumSum

    def test_order(self):
        v = pd.Series([1, 2, 2])
        g = pd.Series(["a", "b", "a"])

        answers = ([1, 3], [2])

        function = self.primitive().get_function()
        for (_, group), answer in zip(v.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)

    def test_regular(self):
        v = pd.Series([101, 102, 103, 104, 105, 106])
        g = pd.Series(["a", "b", "a", "c", "d", "b"])
        answers = ([101, 204], [102, 208], [104], [105])

        function = self.primitive().get_function()
        for (_, group), answer in zip(v.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)


class TestCumMean:
    primitive = CumMean

    def test_order(self):
        v = pd.Series([1, 2, 2])
        g = pd.Series(["a", "b", "a"])

        answers = ([1, 1.5], [2])

        function = self.primitive().get_function()
        for (_, group), answer in zip(v.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)

    def test_regular(self):
        v = pd.Series([101, 102, 103, 104, 105, 106])
        g = pd.Series(["a", "b", "a", "c", "d", "b"])
        answers = ([101, 102], [102, 104], [104], [105])

        function = self.primitive().get_function()
        for (_, group), answer in zip(v.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)


class TestCumMax:
    primitive = CumMax

    def test_order(self):
        v = pd.Series([1, 2, 2])
        g = pd.Series(["a", "b", "a"])

        answers = ([1, 2], [2])

        function = self.primitive().get_function()
        for (_, group), answer in zip(v.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)

    def test_regular(self):
        v = pd.Series([101, 102, 103, 104, 105, 106])
        g = pd.Series(["a", "b", "a", "c", "d", "b"])
        answers = ([101, 103], [102, 106], [104], [105])

        function = self.primitive().get_function()
        for (_, group), answer in zip(v.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)


class TestCumMin:
    primitive = CumMin

    def test_order(self):
        v = pd.Series([1, 2, 2])
        g = pd.Series(["a", "b", "a"])

        answers = ([1, 1], [2])

        function = self.primitive().get_function()
        for (_, group), answer in zip(v.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)

    def test_regular(self):
        v = pd.Series([101, 102, 103, 104, 105, 106, 100])
        g = pd.Series(["a", "b", "a", "c", "d", "b", "a"])
        answers = ([101, 101, 100], [102, 102], [104], [105])

        function = self.primitive().get_function()
        for (_, group), answer in zip(v.groupby(g), answers):
            np.testing.assert_array_equal(function(group), answer)


def test_cum_sum(pd_es):
    log_value_feat = IdentityFeature(pd_es["log"].ww["value"])
    dfeat = Feature(
        IdentityFeature(pd_es["sessions"].ww["device_type"]),
        dataframe_name="log",
    )
    cum_sum = Feature(log_value_feat, groupby=dfeat, primitive=CumSum)
    features = [cum_sum]
    df = calculate_feature_matrix(
        entityset=pd_es,
        features=features,
        instance_ids=range(15),
    )
    cvalues = df[cum_sum.get_name()].values
    assert len(cvalues) == 15
    cum_sum_values = [0, 5, 15, 30, 50, 0, 1, 3, 6, 6, 50, 55, 55, 62, 76]
    for i, v in enumerate(cum_sum_values):
        assert v == cvalues[i]


def test_cum_min(pd_es):
    log_value_feat = IdentityFeature(pd_es["log"].ww["value"])
    cum_min = Feature(
        log_value_feat,
        groupby=IdentityFeature(pd_es["log"].ww["session_id"]),
        primitive=CumMin,
    )
    features = [cum_min]
    df = calculate_feature_matrix(
        entityset=pd_es,
        features=features,
        instance_ids=range(15),
    )
    cvalues = df[cum_min.get_name()].values
    assert len(cvalues) == 15
    cum_min_values = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    for i, v in enumerate(cum_min_values):
        assert v == cvalues[i]


def test_cum_max(pd_es):
    log_value_feat = IdentityFeature(pd_es["log"].ww["value"])
    cum_max = Feature(
        log_value_feat,
        groupby=IdentityFeature(pd_es["log"].ww["session_id"]),
        primitive=CumMax,
    )
    features = [cum_max]
    df = calculate_feature_matrix(
        entityset=pd_es,
        features=features,
        instance_ids=range(15),
    )
    cvalues = df[cum_max.get_name()].values
    assert len(cvalues) == 15
    cum_max_values = [0, 5, 10, 15, 20, 0, 1, 2, 3, 0, 0, 5, 0, 7, 14]
    for i, v in enumerate(cum_max_values):
        assert v == cvalues[i]


def test_cum_sum_group_on_nan(pd_es):
    log_value_feat = IdentityFeature(pd_es["log"].ww["value"])
    pd_es["log"]["product_id"] = (
        ["coke zero"] * 3
        + ["car"] * 2
        + ["toothpaste"] * 3
        + ["brown bag"] * 2
        + ["shoes"]
        + [np.nan] * 4
        + ["coke_zero"] * 2
    )
    pd_es["log"]["value"][16] = 10
    cum_sum = Feature(
        log_value_feat,
        groupby=IdentityFeature(pd_es["log"].ww["product_id"]),
        primitive=CumSum,
    )
    features = [cum_sum]
    df = calculate_feature_matrix(
        entityset=pd_es,
        features=features,
        instance_ids=range(17),
    )
    cvalues = df[cum_sum.get_name()].values
    assert len(cvalues) == 17
    cum_sum_values = [
        0,
        5,
        15,
        15,
        35,
        0,
        1,
        3,
        3,
        3,
        0,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        10,
    ]

    assert len(cvalues) == len(cum_sum_values)
    for i, v in enumerate(cum_sum_values):
        if np.isnan(v):
            assert np.isnan(cvalues[i])
        else:
            assert v == cvalues[i]


def test_cum_sum_numpy_group_on_nan(pd_es):
    class CumSumNumpy(TransformPrimitive):
        """Returns the cumulative sum after grouping"""

        name = "cum_sum"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        uses_full_dataframe = True

        def get_function(self):
            def cum_sum(values):
                return values.cumsum().values

            return cum_sum

    log_value_feat = IdentityFeature(pd_es["log"].ww["value"])
    pd_es["log"]["product_id"] = (
        ["coke zero"] * 3
        + ["car"] * 2
        + ["toothpaste"] * 3
        + ["brown bag"] * 2
        + ["shoes"]
        + [np.nan] * 4
        + ["coke_zero"] * 2
    )
    pd_es["log"]["value"][16] = 10
    cum_sum = Feature(
        log_value_feat,
        groupby=IdentityFeature(pd_es["log"].ww["product_id"]),
        primitive=CumSumNumpy,
    )
    assert cum_sum.get_name() == "CUM_SUM(value) by product_id"
    features = [cum_sum]
    df = calculate_feature_matrix(
        entityset=pd_es,
        features=features,
        instance_ids=range(17),
    )
    cvalues = df[cum_sum.get_name()].values
    assert len(cvalues) == 17
    cum_sum_values = [
        0,
        5,
        15,
        15,
        35,
        0,
        1,
        3,
        3,
        3,
        0,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        10,
    ]

    assert len(cvalues) == len(cum_sum_values)
    for i, v in enumerate(cum_sum_values):
        if np.isnan(v):
            assert np.isnan(cvalues[i])
        else:
            assert v == cvalues[i]


def test_cum_handles_uses_full_dataframe(pd_es):
    def check(feature):
        feature_set = FeatureSet([feature])
        calculator = FeatureSetCalculator(
            pd_es,
            feature_set=feature_set,
            time_last=None,
        )
        df_1 = calculator.run(np.array([0, 1, 2]))
        df_2 = calculator.run(np.array([2, 4]))

        # check that the value for instance id 2 matches
        assert (df_2.loc[2] == df_1.loc[2]).all()

    for primitive in [CumSum, CumMean, CumMax, CumMin]:
        check(
            Feature(
                pd_es["log"].ww["value"],
                groupby=IdentityFeature(pd_es["log"].ww["session_id"]),
                primitive=primitive,
            ),
        )

    check(
        Feature(
            pd_es["log"].ww["product_id"],
            groupby=Feature(pd_es["log"].ww["product_id"]),
            primitive=CumCount,
        ),
    )


def test_cum_mean(pd_es):
    log_value_feat = IdentityFeature(pd_es["log"].ww["value"])
    cum_mean = Feature(
        log_value_feat,
        groupby=IdentityFeature(pd_es["log"].ww["session_id"]),
        primitive=CumMean,
    )
    features = [cum_mean]
    df = calculate_feature_matrix(
        entityset=pd_es,
        features=features,
        instance_ids=range(15),
    )
    cvalues = df[cum_mean.get_name()].values
    assert len(cvalues) == 15
    cum_mean_values = [0, 2.5, 5, 7.5, 10, 0, 0.5, 1, 1.5, 0, 0, 2.5, 0, 3.5, 7]
    for i, v in enumerate(cum_mean_values):
        assert v == cvalues[i]


def test_cum_count(pd_es):
    cum_count = Feature(
        IdentityFeature(pd_es["log"].ww["product_id"]),
        groupby=IdentityFeature(pd_es["log"].ww["product_id"]),
        primitive=CumCount,
    )
    features = [cum_count]
    df = calculate_feature_matrix(
        entityset=pd_es,
        features=features,
        instance_ids=range(15),
    )
    cvalues = df[cum_count.get_name()].values
    assert len(cvalues) == 15
    cum_count_values = [1, 2, 3, 1, 2, 1, 2, 3, 1, 2, 1, 4, 5, 6, 7]
    for i, v in enumerate(cum_count_values):
        assert v == cvalues[i]


def test_rename(pd_es):
    cum_count = Feature(
        IdentityFeature(pd_es["log"].ww["product_id"]),
        groupby=IdentityFeature(pd_es["log"].ww["product_id"]),
        primitive=CumCount,
    )
    copy_feat = cum_count.rename("rename_test")
    assert cum_count.unique_name() != copy_feat.unique_name()
    assert cum_count.get_name() != copy_feat.get_name()
    assert all(
        [
            x.generate_name() == y.generate_name()
            for x, y in zip(cum_count.base_features, copy_feat.base_features)
        ],
    )
    assert cum_count.dataframe_name == copy_feat.dataframe_name


def test_groupby_no_data(pd_es):
    cum_count = Feature(
        IdentityFeature(pd_es["log"].ww["product_id"]),
        groupby=IdentityFeature(pd_es["log"].ww["product_id"]),
        primitive=CumCount,
    )
    last_feat = Feature(cum_count, parent_dataframe_name="customers", primitive=Last)
    df = calculate_feature_matrix(
        entityset=pd_es,
        features=[last_feat],
        cutoff_time=pd.Timestamp("2011-04-08"),
    )
    cvalues = df[last_feat.get_name()].values
    assert len(cvalues) == 2
    assert all([pd.isnull(value) for value in cvalues])


def test_groupby_uses_calc_time(pd_es):
    def projected_amount_left(amount, timestamp, time=None):
        # cumulative sum of amount, with timedelta *  constant subtracted
        delta = time - timestamp
        delta_seconds = delta / np.timedelta64(1, "s")
        return amount.cumsum() - (delta_seconds)

    class ProjectedAmountRemaining(TransformPrimitive):
        name = "projected_amount_remaining"
        uses_calc_time = True
        input_types = [
            ColumnSchema(semantic_tags={"numeric"}),
            ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"}),
        ]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        uses_full_dataframe = True

        def get_function(self):
            return projected_amount_left

    time_since_product = GroupByTransformFeature(
        [
            IdentityFeature(pd_es["log"].ww["value"]),
            IdentityFeature(pd_es["log"].ww["datetime"]),
        ],
        groupby=IdentityFeature(pd_es["log"].ww["product_id"]),
        primitive=ProjectedAmountRemaining,
    )
    df = calculate_feature_matrix(
        entityset=pd_es,
        features=[time_since_product],
        cutoff_time=pd.Timestamp("2011-04-10 11:10:30"),
    )
    answers = [
        -88830,
        -88819,
        -88803,
        -88797,
        -88771,
        -88770,
        -88760,
        -88749,
        -88740,
        -88227,
        -1830,
        -1809,
        -1750,
        -1740,
        -1723,
        np.nan,
        np.nan,
    ]

    for x, y in zip(df[time_since_product.get_name()], answers):
        assert (pd.isnull(x) and pd.isnull(y)) or x == y


def test_groupby_multi_output_stacking(pd_es):
    class TestTime(TransformPrimitive):
        name = "test_time"
        input_types = [ColumnSchema(logical_type=Datetime)]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        number_output_features = 6

    fl = dfs(
        entityset=pd_es,
        target_dataframe_name="sessions",
        agg_primitives=["sum"],
        groupby_trans_primitives=[TestTime],
        features_only=True,
        max_depth=4,
    )

    for i in range(6):
        f = "SUM(log.TEST_TIME(datetime)[%d] by product_id)" % i
        assert feature_with_name(fl, f)
        assert ("customers.SUM(log.TEST_TIME(datetime)[%d] by session_id)" % i) in fl


def test_serialization(pd_es):
    value = IdentityFeature(pd_es["log"].ww["value"])
    zipcode = IdentityFeature(pd_es["log"].ww["zipcode"])
    primitive = CumSum()
    groupby = feature_base.GroupByTransformFeature(value, primitive, zipcode)

    dictionary = {
        "name": "CUM_SUM(value) by zipcode",
        "base_features": [value.unique_name()],
        "primitive": primitive,
        "groupby": zipcode.unique_name(),
    }

    assert dictionary == groupby.get_arguments()
    dependencies = {
        value.unique_name(): value,
        zipcode.unique_name(): zipcode,
    }
    assert groupby == feature_base.GroupByTransformFeature.from_dictionary(
        dictionary,
        pd_es,
        dependencies,
        primitive,
    )


def test_groupby_with_multioutput_primitive(pd_es):
    class MultiCumSum(TransformPrimitive):
        name = "multi_cum_sum"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        number_output_features = 3

        def get_function(self):
            def multi_cum_sum(x):
                return x.cumsum(), x.cummax(), x.cummin()

            return multi_cum_sum

    fm, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="customers",
        trans_primitives=[],
        agg_primitives=[],
        groupby_trans_primitives=[MultiCumSum, CumSum, CumMax, CumMin],
    )

    # Calculate output in a separate DFS call to make sure the multi-output code
    # does not alter any values
    fm2, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="customers",
        trans_primitives=[],
        agg_primitives=[],
        groupby_trans_primitives=[CumSum, CumMax, CumMin],
    )

    answer_cols = [
        ["CUM_SUM(age) by cohort", "CUM_SUM(age) by région_id"],
        ["CUM_MAX(age) by cohort", "CUM_MAX(age) by région_id"],
        ["CUM_MIN(age) by cohort", "CUM_MIN(age) by région_id"],
    ]

    for i in range(3):
        # Check that multi-output gives correct answers
        f = "MULTI_CUM_SUM(age)[%d] by cohort" % i
        assert f in fm.columns
        for x, y in zip(fm[f].values, fm[answer_cols[i][0]].values):
            assert x == y
        f = "MULTI_CUM_SUM(age)[%d] by région_id" % i
        assert f in fm.columns
        for x, y in zip(fm[f].values, fm[answer_cols[i][1]].values):
            assert x == y
        # Verify single output results are unchanged by inclusion of
        # multi-output primitive
        for x, y in zip(fm[answer_cols[i][0]], fm2[answer_cols[i][0]]):
            assert x == y
        for x, y in zip(fm[answer_cols[i][1]], fm2[answer_cols[i][1]]):
            assert x == y


def test_groupby_with_multioutput_primitive_custom_names(pd_es):
    class MultiCumSum(TransformPrimitive):
        name = "multi_cum_sum"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        number_output_features = 3

        def get_function(self):
            def multi_cum_sum(x):
                return x.cumsum(), x.cummax(), x.cummin()

            return multi_cum_sum

        def generate_names(primitive, base_feature_names):
            return ["CUSTOM_SUM", "CUSTOM_MAX", "CUSTOM_MIN"]

    fm, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="customers",
        trans_primitives=[],
        agg_primitives=[],
        groupby_trans_primitives=[MultiCumSum, CumSum, CumMax, CumMin],
    )

    answer_cols = [
        ["CUM_SUM(age) by cohort", "CUM_SUM(age) by région_id"],
        ["CUM_MAX(age) by cohort", "CUM_MAX(age) by région_id"],
        ["CUM_MIN(age) by cohort", "CUM_MIN(age) by région_id"],
    ]

    expected_names = [
        ["CUSTOM_SUM by cohort", "CUSTOM_SUM by région_id"],
        ["CUSTOM_MAX by cohort", "CUSTOM_MAX by région_id"],
        ["CUSTOM_MIN by cohort", "CUSTOM_MIN by région_id"],
    ]

    for i in range(3):
        f = expected_names[i][0]
        assert f in fm.columns
        for x, y in zip(fm[f].values, fm[answer_cols[i][0]].values):
            assert x == y
        f = expected_names[i][1]
        assert f in fm.columns
        for x, y in zip(fm[f].values, fm[answer_cols[i][1]].values):
            assert x == y



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable

from featuretools.primitives.base import TransformPrimitive


class CumCount(TransformPrimitive):
    """Calculates the cumulative count.

    Description:
        Given a list of values, return the cumulative count
        (or running count). There is no set window, so the
        count at each point is calculated over all prior
        values. `NaN` values are counted.

    Examples:
        >>> cum_count = CumCount()
        >>> cum_count([1, 2, 3, 4, None, 5]).tolist()
        [1, 2, 3, 4, 5, 6]
    """

    name = "cum_count"
    input_types = [
        [ColumnSchema(semantic_tags={"foreign_key"})],
        [ColumnSchema(semantic_tags={"category"})],
    ]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    uses_full_dataframe = True
    description_template = "the cumulative count of {}"

    def get_function(self):
        def cum_count(values):
            return np.arange(1, len(values) + 1)

        return cum_count



import re
from string import punctuation
from typing import Optional

import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable, NaturalLanguage

from featuretools.primitives.base import TransformPrimitive
from featuretools.primitives.standard.transform.natural_language.constants import (
    DELIMITERS,
)
from featuretools.utils.gen_utils import Library


class NumWords(TransformPrimitive):
    """Determines the number of words in a string. Words are sequences of characters
    delimited by whitespace.

    Examples:
        >>> num_words = NumWords()
        >>> num_words(['This is a string',
        ...            'Two words',
        ...            'no-spaces',
        ...            'Also works with sentences. Second sentence!']).tolist()
        [4, 2, 1, 6]
    """

    name = "num_words"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the number of words in {}"

    def get_function(self):
        def word_counter(array):
            def _get_number_of_words(elem: Optional[str]):
                """Returns the number of words in given element,
                or pd.NA given null input"""
                if pd.isna(elem):
                    return pd.NA
                return sum(
                    1 for word in re.split(DELIMITERS, elem) if word.strip(punctuation)
                )

            return array.apply(_get_number_of_words)

        return word_counter



import os
import re

import graphviz
import pandas as pd
import pytest

from featuretools import EntitySet
from featuretools.utils.gen_utils import Library


@pytest.fixture
def pd_simple():
    es = EntitySet("test")
    df = pd.DataFrame({"foo": [1]})
    es.add_dataframe(df, dataframe_name="test", index="foo")
    return es


@pytest.fixture
def dd_simple():
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    es = EntitySet("test")
    df = pd.DataFrame({"foo": [1]})
    df = dd.from_pandas(df, npartitions=2)
    es.add_dataframe(df, dataframe_name="test", index="foo")
    return es


@pytest.fixture
def spark_simple():
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    es = EntitySet("test")
    df = ps.DataFrame({"foo": [1]})
    es.add_dataframe(df, dataframe_name="test", index="foo")
    return es


@pytest.fixture(params=["pd_simple", "dd_simple", "spark_simple"])
def simple_es(request):
    return request.getfixturevalue(request.param)


def test_returns_digraph_object(es):
    graph = es.plot()

    assert isinstance(graph, graphviz.Digraph)


def test_saving_png_file(es, tmp_path):
    output_path = str(tmp_path.joinpath("test1.png"))

    es.plot(to_file=output_path)

    assert os.path.isfile(output_path)
    os.remove(output_path)


def test_missing_file_extension(es):
    output_path = "test1"

    with pytest.raises(ValueError) as excinfo:
        es.plot(to_file=output_path)

    assert str(excinfo.value).startswith("Please use a file extension")


def test_invalid_format(es):
    output_path = "test1.xzy"

    with pytest.raises(ValueError) as excinfo:
        es.plot(to_file=output_path)

    assert str(excinfo.value).startswith("Unknown format")


def test_multiple_rows(es):
    plot_ = es.plot()
    result = re.findall(r"\((\d+\srows?)\)", plot_.source)
    expected = ["{} rows".format(str(i.shape[0])) for i in es.dataframes]
    if es.dataframe_type == Library.DASK:
        # Dask does not list number of rows in plot
        assert result == []
    else:
        assert result == expected


def test_single_row(simple_es):
    plot_ = simple_es.plot()
    result = re.findall(r"\((\d+\srows?)\)", plot_.source)
    expected = ["1 row"]
    if simple_es.dataframe_type == Library.DASK:
        # Dask does not list number of rows in plot
        assert result == []
    else:
        assert result == expected



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable, NaturalLanguage

from featuretools.primitives.standard.transform.natural_language.count_string import (
    CountString,
)


class NumberOfHashtags(CountString):
    """Determines the number of hashtags in a string.

    Description:
        Given a list of strings, determine the number of hashtags
        in each string.

        A hashtag is defined as a string that meets the following criteria:
            - Starts with a '#' character, followed by a sequence of alphanumeric characters containing at least one alphabetic character
            - Present at the start of a string or after whitespace
            - Terminated by the end of the string, a whitespace, or a punctuation character other than '#'
                - e.g. The string '#yes-no' contains a valid hashtag ('#yes')
                - e.g. The string '#yes#' does not contain a valid hashtag

        This implementation handles Unicode characters.

        This implementation does not impose any character limit on hashtags.

        If a string is missing, return `NaN`.

    Examples:
        >>> x = ['#regular #expression', 'this is a string', '###__regular#1and_0#expression']
        >>> number_of_hashtags = NumberOfHashtags()
        >>> number_of_hashtags(x).tolist()
        [2.0, 0.0, 0.0]
    """

    name = "number_of_hashtags"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    default_value = 0

    def __init__(self):
        pattern = r"((^#)|\s#)(\w*([^\W\d])+\w*)(?![#\w])"
        super().__init__(string=pattern, is_regex=True, ignore_case=False)



from featuretools.primitives.standard.transform.url.url_to_domain import URLToDomain
from featuretools.primitives.standard.transform.url.url_to_protocol import URLToProtocol
from featuretools.primitives.standard.transform.url.url_to_tld import URLToTLD



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class LessThanEqualToScalar(TransformPrimitive):
    """Determines if values are less than or equal to a given scalar.

    Description:
        Given a list of values and a constant scalar, determine
        whether each of the values is less than or equal to the
        scalar. If a value is equal to the scalar, return `True`.

    Examples:
        >>> less_than_equal_to_scalar = LessThanEqualToScalar(value=2)
        >>> less_than_equal_to_scalar([3, 1, 2]).tolist()
        [False, True, True]
    """

    name = "less_than_equal_to_scalar"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]

    def __init__(self, value=0):
        self.value = value
        self.description_template = "whether {{}} is less than or equal to {}".format(
            self.value,
        )

    def get_function(self):
        def less_than_equal_to_scalar(vals):
            return vals <= self.value

        return less_than_equal_to_scalar

    def generate_name(self, base_feature_names):
        return "%s <= %s" % (base_feature_names[0], str(self.value))



import os.path
import re

import pytest
from pympler.asizeof import asizeof
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Integer

from featuretools import Feature, config, feature_base
from featuretools.feature_base import IdentityFeature
from featuretools.primitives import (
    Count,
    Diff,
    Last,
    Mode,
    Negate,
    NMostCommon,
    NumUnique,
    Sum,
    TransformPrimitive,
)
from featuretools.synthesis.deep_feature_synthesis import can_stack_primitive_on_inputs
from featuretools.tests.testing_utils import check_rename


def test_copy_features_does_not_copy_entityset(es):
    agg = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        primitive=Sum,
    )
    agg_where = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        where=IdentityFeature(es["log"].ww["value"]) == 2,
        primitive=Sum,
    )
    agg_use_previous = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        use_previous="4 days",
        primitive=Sum,
    )
    agg_use_previous_where = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        where=IdentityFeature(es["log"].ww["value"]) == 2,
        use_previous="4 days",
        primitive=Sum,
    )
    features = [agg, agg_where, agg_use_previous, agg_use_previous_where]
    in_memory_size = asizeof(locals())
    copied = [f.copy() for f in features]
    new_in_memory_size = asizeof(locals())
    assert new_in_memory_size < 2 * in_memory_size


def test_get_dependencies(es):
    f = Feature(es["log"].ww["value"])
    agg1 = Feature(f, parent_dataframe_name="sessions", primitive=Sum)
    agg2 = Feature(agg1, parent_dataframe_name="customers", primitive=Sum)
    d1 = Feature(agg2, "sessions")
    shallow = d1.get_dependencies(deep=False, ignored=None)
    deep = d1.get_dependencies(deep=True, ignored=None)
    ignored = set([agg1.unique_name()])
    deep_ignored = d1.get_dependencies(deep=True, ignored=ignored)
    assert [s.unique_name() for s in shallow] == [agg2.unique_name()]
    assert [d.unique_name() for d in deep] == [
        agg2.unique_name(),
        agg1.unique_name(),
        f.unique_name(),
    ]
    assert [d.unique_name() for d in deep_ignored] == [agg2.unique_name()]


def test_get_depth(es):
    f = Feature(es["log"].ww["value"])
    g = Feature(es["log"].ww["value"])
    agg1 = Feature(f, parent_dataframe_name="sessions", primitive=Last)
    agg2 = Feature(agg1, parent_dataframe_name="customers", primitive=Last)
    d1 = Feature(agg2, "sessions")
    d2 = Feature(d1, "log")
    assert d2.get_depth() == 4
    # Make sure this works if we pass in two of the same
    # feature. This came up when user supplied duplicates
    # in the seed_features of DFS.
    assert d2.get_depth(stop_at=[f, g]) == 4
    assert d2.get_depth(stop_at=[f, g, agg1]) == 3
    assert d2.get_depth(stop_at=[f, g, agg1]) == 3
    assert d2.get_depth(stop_at=[f, g, agg2]) == 2
    assert d2.get_depth(stop_at=[f, g, d1]) == 1
    assert d2.get_depth(stop_at=[f, g, d2]) == 0


def test_squared(es):
    feature = Feature(es["log"].ww["value"])
    squared = feature * feature
    assert len(squared.base_features) == 2
    assert (
        squared.base_features[0].unique_name() == squared.base_features[1].unique_name()
    )


def test_return_type_inference(es):
    mode = Feature(
        es["log"].ww["priority_level"],
        parent_dataframe_name="customers",
        primitive=Mode,
    )
    assert (
        mode.column_schema
        == IdentityFeature(es["log"].ww["priority_level"]).column_schema
    )


def test_return_type_inference_direct_feature(es):
    mode = Feature(
        es["log"].ww["priority_level"],
        parent_dataframe_name="customers",
        primitive=Mode,
    )
    mode_session = Feature(mode, "sessions")
    assert (
        mode_session.column_schema
        == IdentityFeature(es["log"].ww["priority_level"]).column_schema
    )


def test_return_type_inference_index(es):
    last = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="customers",
        primitive=Last,
    )
    assert "index" not in last.column_schema.semantic_tags
    assert isinstance(last.column_schema.logical_type, Integer)


def test_return_type_inference_datetime_time_index(es):
    last = Feature(
        es["log"].ww["datetime"],
        parent_dataframe_name="customers",
        primitive=Last,
    )
    assert isinstance(last.column_schema.logical_type, Datetime)


def test_return_type_inference_numeric_time_index(int_es):
    last = Feature(
        int_es["log"].ww["datetime"],
        parent_dataframe_name="customers",
        primitive=Last,
    )
    assert "numeric" in last.column_schema.semantic_tags


def test_return_type_inference_id(es):
    # direct features should keep foreign key tag
    direct_id_feature = Feature(es["sessions"].ww["customer_id"], "log")
    assert "foreign_key" in direct_id_feature.column_schema.semantic_tags

    # aggregations of foreign key types should get converted
    last_feat = Feature(
        es["log"].ww["session_id"],
        parent_dataframe_name="customers",
        primitive=Last,
    )
    assert "foreign_key" not in last_feat.column_schema.semantic_tags
    assert isinstance(last_feat.column_schema.logical_type, Integer)

    # also test direct feature of aggregation
    last_direct = Feature(last_feat, "sessions")
    assert "foreign_key" not in last_direct.column_schema.semantic_tags
    assert isinstance(last_direct.column_schema.logical_type, Integer)


def test_set_data_path(es):
    key = "primitive_data_folder"

    # Don't change orig_path
    orig_path = config.get(key)
    new_path = "/example/new/directory"
    filename = "test.csv"

    # Test that default path works
    sum_prim = Sum()
    assert sum_prim.get_filepath(filename) == os.path.join(orig_path, filename)

    # Test that new path works
    config.set({key: new_path})
    assert sum_prim.get_filepath(filename) == os.path.join(new_path, filename)

    # Test that new path with trailing / works
    new_path += "/"
    config.set({key: new_path})
    assert sum_prim.get_filepath(filename) == os.path.join(new_path, filename)

    # Test that the path is correct on newly defined feature
    sum_prim2 = Sum()
    assert sum_prim2.get_filepath(filename) == os.path.join(new_path, filename)

    # Ensure path was reset
    config.set({key: orig_path})
    assert config.get(key) == orig_path


def test_to_dictionary_direct(es):
    actual = Feature(
        IdentityFeature(es["sessions"].ww["customer_id"]),
        "log",
    ).to_dictionary()

    expected = {
        "type": "DirectFeature",
        "dependencies": ["sessions: customer_id"],
        "arguments": {
            "name": "sessions.customer_id",
            "base_feature": "sessions: customer_id",
            "relationship": {
                "parent_dataframe_name": "sessions",
                "child_dataframe_name": "log",
                "parent_column_name": "id",
                "child_column_name": "session_id",
            },
        },
    }

    assert expected == actual


def test_to_dictionary_identity(es):
    actual = Feature(es["sessions"].ww["customer_id"]).to_dictionary()

    expected = {
        "type": "IdentityFeature",
        "dependencies": [],
        "arguments": {
            "name": "customer_id",
            "column_name": "customer_id",
            "dataframe_name": "sessions",
        },
    }

    assert expected == actual


def test_to_dictionary_agg(es):
    primitive = Sum()
    actual = Feature(
        es["customers"].ww["age"],
        primitive=primitive,
        parent_dataframe_name="cohorts",
    ).to_dictionary()

    expected = {
        "type": "AggregationFeature",
        "dependencies": ["customers: age"],
        "arguments": {
            "name": "SUM(customers.age)",
            "base_features": ["customers: age"],
            "relationship_path": [
                {
                    "parent_dataframe_name": "cohorts",
                    "child_dataframe_name": "customers",
                    "parent_column_name": "cohort",
                    "child_column_name": "cohort",
                },
            ],
            "primitive": primitive,
            "where": None,
            "use_previous": None,
        },
    }

    assert expected == actual


def test_to_dictionary_where(es):
    primitive = Sum()
    actual = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        where=IdentityFeature(es["log"].ww["value"]) == 2,
        primitive=primitive,
    ).to_dictionary()

    expected = {
        "type": "AggregationFeature",
        "dependencies": ["log: value", "log: value = 2"],
        "arguments": {
            "name": "SUM(log.value WHERE value = 2)",
            "base_features": ["log: value"],
            "relationship_path": [
                {
                    "parent_dataframe_name": "sessions",
                    "child_dataframe_name": "log",
                    "parent_column_name": "id",
                    "child_column_name": "session_id",
                },
            ],
            "primitive": primitive,
            "where": "log: value = 2",
            "use_previous": None,
        },
    }

    assert expected == actual


def test_to_dictionary_trans(es):
    primitive = Negate()
    trans_feature = Feature(es["customers"].ww["age"], primitive=primitive)

    expected = {
        "type": "TransformFeature",
        "dependencies": ["customers: age"],
        "arguments": {
            "name": "-(age)",
            "base_features": ["customers: age"],
            "primitive": primitive,
        },
    }

    assert expected == trans_feature.to_dictionary()


def test_to_dictionary_groupby_trans(es):
    primitive = Negate()
    id_feat = Feature(es["log"].ww["product_id"])
    groupby_feature = Feature(
        es["log"].ww["value"],
        primitive=primitive,
        groupby=id_feat,
    )

    expected = {
        "type": "GroupByTransformFeature",
        "dependencies": ["log: value", "log: product_id"],
        "arguments": {
            "name": "-(value) by product_id",
            "base_features": ["log: value"],
            "primitive": primitive,
            "groupby": "log: product_id",
        },
    }

    assert expected == groupby_feature.to_dictionary()


def test_to_dictionary_multi_slice(es):
    slice_feature = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )[0]

    expected = {
        "type": "FeatureOutputSlice",
        "dependencies": ["customers: N_MOST_COMMON(log.product_id, n=2)"],
        "arguments": {
            "name": "N_MOST_COMMON(log.product_id, n=2)[0]",
            "base_feature": "customers: N_MOST_COMMON(log.product_id, n=2)",
            "n": 0,
        },
    }

    assert expected == slice_feature.to_dictionary()


def test_multi_output_base_error_agg(es):
    three_common = NMostCommon(3)
    tc = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="sessions",
        primitive=three_common,
    )
    error_text = "Cannot stack on whole multi-output feature."
    with pytest.raises(ValueError, match=error_text):
        Feature(tc, parent_dataframe_name="customers", primitive=NumUnique)


def test_multi_output_base_error_trans(es):
    class TestTime(TransformPrimitive):
        name = "test_time"
        input_types = [ColumnSchema(logical_type=Datetime)]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        number_output_features = 6

    tc = Feature(es["customers"].ww["birthday"], primitive=TestTime)

    error_text = "Cannot stack on whole multi-output feature."
    with pytest.raises(ValueError, match=error_text):
        Feature(tc, primitive=Diff)


def test_multi_output_attributes(es):
    tc = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="sessions",
        primitive=NMostCommon,
    )

    assert tc.generate_name() == "N_MOST_COMMON(log.product_id)"
    assert tc.number_output_features == 3
    assert tc.base_features == ["<Feature: product_id>"]

    assert tc[0].generate_name() == "N_MOST_COMMON(log.product_id)[0]"
    assert tc[0].number_output_features == 1
    assert tc[0].base_features == [tc]
    assert tc.relationship_path == tc[0].relationship_path


def test_multi_output_index_error(es):
    error_text = "can only access slice of multi-output feature"
    three_common = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="sessions",
        primitive=NMostCommon,
    )

    with pytest.raises(AssertionError, match=error_text):
        single = Feature(
            es["log"].ww["product_id"],
            parent_dataframe_name="sessions",
            primitive=NumUnique,
        )
        single[0]

    error_text = "Cannot get item from slice of multi output feature"
    with pytest.raises(ValueError, match=error_text):
        three_common[0][0]

    error_text = "index is higher than the number of outputs"
    with pytest.raises(AssertionError, match=error_text):
        three_common[10]


def test_rename(es):
    feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )
    new_name = "session_test"
    new_names = ["session_test"]
    check_rename(feat, new_name, new_names)


def test_rename_multioutput(es):
    feat = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    new_name = "session_test"
    new_names = ["session_test[0]", "session_test[1]"]
    check_rename(feat, new_name, new_names)


def test_rename_featureoutputslice(es):
    multi_output_feat = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    feat = feature_base.FeatureOutputSlice(multi_output_feat, 0)
    new_name = "session_test"
    new_names = ["session_test"]
    check_rename(feat, new_name, new_names)


def test_set_feature_names_wrong_number_of_names(es):
    feat = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    new_names = ["col1"]
    error_msg = re.escape(
        "Number of names provided must match the number of output features: 1 name(s) provided, 2 expected.",
    )
    with pytest.raises(ValueError, match=error_msg):
        feat.set_feature_names(new_names)


def test_set_feature_names_not_unique(es):
    feat = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    new_names = ["col1", "col1"]
    error_msg = "Provided output feature names must be unique."
    with pytest.raises(ValueError, match=error_msg):
        feat.set_feature_names(new_names)


def test_set_feature_names_error_on_single_output_feature(es):
    feat = Feature(es["sessions"].ww["device_name"], "log")
    new_names = ["sessions_device"]
    error_msg = "The set_feature_names can only be used on features that have more than one output column."
    with pytest.raises(ValueError, match=error_msg):
        feat.set_feature_names(new_names)


def test_set_feature_names_transform_feature(es):
    class MultiCumulative(TransformPrimitive):
        name = "multi_cum_sum"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        number_output_features = 3

    feat = Feature(es["log"].ww["value"], primitive=MultiCumulative)
    new_names = ["cumulative_sum", "cumulative_max", "cumulative_min"]
    feat.set_feature_names(new_names)
    assert feat.get_feature_names() == new_names


def test_set_feature_names_aggregation_feature(es):
    feat = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    new_names = ["agg_col_1", "second_agg_col"]
    feat.set_feature_names(new_names)
    assert feat.get_feature_names() == new_names


def test_renaming_resets_feature_output_names_to_default(es):
    feat = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    new_names = ["renamed1", "renamed2"]
    feat.set_feature_names(new_names)
    assert feat.get_feature_names() == new_names

    feat = feat.rename("new_feature_name")
    assert feat.get_feature_names() == ["new_feature_name[0]", "new_feature_name[1]"]


def test_base_of_and_stack_on_heuristic(es, test_aggregation_primitive):
    child = Feature(
        es["sessions"].ww["id"],
        parent_dataframe_name="customers",
        primitive=Count,
    )
    test_aggregation_primitive.stack_on = []
    child.primitive.base_of = []
    assert not can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on = []
    child.primitive.base_of = None
    assert can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on = []
    child.primitive.base_of = [test_aggregation_primitive]
    assert can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on = None
    child.primitive.base_of = []
    assert can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on = None
    child.primitive.base_of = None
    assert can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on = None
    child.primitive.base_of = [test_aggregation_primitive]
    assert can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on = [type(child.primitive)]
    child.primitive.base_of = []
    assert can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on = [type(child.primitive)]
    child.primitive.base_of = None
    assert can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on = [type(child.primitive)]
    child.primitive.base_of = [test_aggregation_primitive]
    assert can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on = None
    child.primitive.base_of = None
    child.primitive.base_of_exclude = [test_aggregation_primitive]
    assert not can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    test_aggregation_primitive.stack_on_exclude = [Count]
    assert not can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])

    child.primitive.number_output_features = 2
    test_aggregation_primitive.stack_on_exclude = []
    test_aggregation_primitive.stack_on = []
    child.primitive.base_of = []
    assert not can_stack_primitive_on_inputs(test_aggregation_primitive(), [child])


def test_stack_on_self(es, test_transform_primitive):
    # test stacks on self
    child = Feature(
        es["log"].ww["value"],
        primitive=test_transform_primitive,
    )
    test_transform_primitive.stack_on = []
    child.primitive.base_of = []
    test_transform_primitive.stack_on_self = False
    child.primitive.stack_on_self = False
    assert not can_stack_primitive_on_inputs(test_transform_primitive(), [child])

    test_transform_primitive.stack_on_self = True
    assert can_stack_primitive_on_inputs(test_transform_primitive(), [child])

    test_transform_primitive.stack_on = None
    test_transform_primitive.stack_on_self = False
    assert not can_stack_primitive_on_inputs(test_transform_primitive(), [child])



from featuretools.primitives import AggregationPrimitive, TransformPrimitive
from featuretools.primitives.utils import (
    get_aggregation_primitives,
    get_transform_primitives,
)
from featuretools.synthesis.deep_feature_synthesis import DeepFeatureSynthesis
from featuretools.synthesis.utils import _categorize_features, get_unused_primitives
from featuretools.utils.gen_utils import Library


def get_valid_primitives(
    entityset,
    target_dataframe_name,
    max_depth=2,
    selected_primitives=None,
    **dfs_kwargs,
):
    """
    Returns two lists of primitives (transform and aggregation) containing
    primitives that can be applied to the specific target dataframe to create
    features.  If the optional 'selected_primitives' parameter is not used,
    all discoverable primitives will be considered.

    Note:
        When using a ``max_depth`` greater than 1, some primitives returned by
        this function may not create any features if passed to DFS alone.  These
        primitives relied on features created by other primitives as input
        (primitive stacking).

    Args:
        entityset (EntitySet): An already initialized entityset
        target_dataframe_name (str): Name of dataframe to create features for.
        max_depth (int, optional): Maximum allowed depth of features.
        selected_primitives(list[str or AggregationPrimitive/TransformPrimitive], optional):
            list of primitives to consider when looking for valid primitives.
            If None, all primitives will be considered
        dfs_kwargs (keywords): Additional keyword arguments to pass as keyword arguments to
            the DeepFeatureSynthesis object. Should not include ``max_depth``, ``agg_primitives``,
            or ``trans_primitives``, as those are passed in explicity.
    Returns:
       list[AggregationPrimitive], list[TransformPrimitive]:
           The list of valid aggregation primitives and the list of valid
           transform primitives.
    """
    agg_primitives = []
    trans_primitives = []
    available_aggs = get_aggregation_primitives()
    available_trans = get_transform_primitives()

    for library in Library:
        if library.value == entityset.dataframe_type:
            df_library = library
            break

    if selected_primitives:
        for prim in selected_primitives:
            if not isinstance(prim, str):
                if issubclass(prim, AggregationPrimitive):
                    prim_list = agg_primitives
                elif issubclass(prim, TransformPrimitive):
                    prim_list = trans_primitives
                else:
                    raise ValueError(
                        f"Selected primitive {prim} is not an "
                        "AggregationPrimitive, TransformPrimitive, or str",
                    )
            elif prim in available_aggs:
                prim = available_aggs[prim]
                prim_list = agg_primitives
            elif prim in available_trans:
                prim = available_trans[prim]
                prim_list = trans_primitives
            else:
                raise ValueError(f"'{prim}' is not a recognized primitive name")
            if df_library in prim.compatibility:
                prim_list.append(prim)
    else:
        agg_primitives = [
            agg for agg in available_aggs.values() if df_library in agg.compatibility
        ]
        trans_primitives = [
            trans
            for trans in available_trans.values()
            if df_library in trans.compatibility
        ]

    dfs_object = DeepFeatureSynthesis(
        target_dataframe_name,
        entityset,
        agg_primitives=agg_primitives,
        trans_primitives=trans_primitives,
        max_depth=max_depth,
        **dfs_kwargs,
    )

    features = dfs_object.build_features()

    trans, agg, _, _ = _categorize_features(features)

    trans_unused = get_unused_primitives(trans_primitives, trans)
    agg_unused = get_unused_primitives(agg_primitives, agg)

    # switch from str to class
    agg_unused = [available_aggs[name] for name in agg_unused]
    trans_unused = [available_trans[name] for name in trans_unused]

    used_agg_prims = set(agg_primitives).difference(set(agg_unused))
    used_trans_prims = set(trans_primitives).difference(set(trans_unused))
    return list(used_agg_prims), list(used_trans_prims)



import re
from string import punctuation

import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable, NaturalLanguage

from featuretools.primitives.base import TransformPrimitive
from featuretools.primitives.standard.transform.natural_language.constants import (
    DELIMITERS,
)


class NumberOfWordsInQuotes(TransformPrimitive):
    """Determines the number of words in quotes in a string.

    Description:
        Given a list of strings, determine the number of words in quotes
        in each string.

        This implementation handles Unicode characters.

        If a string is missing, return `NaN`.

    Args:
        quote_type (str, optional): Specifies what type of quotation marks to match.
        Argument "single" matches on only single quotes (' ').
        Argument "double" matches words between double quotes (" ").
        Argument "both" matches words between either type of quotes.
        Defaults to "both".

    Examples:
         >>> x = ['"python" java prolog "Diffie-Hellman" "4.99"', "Reach me at 'user@email.com'", "'Here's an interesting example!'"]
        >>> number_of_words_in_quotes = NumberOfWordsInQuotes()
        >>> number_of_words_in_quotes(x).tolist()
        [3, 1, 4]
    """

    name = "number_of_words_in_quotes"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    default_value = 0

    def __init__(self, quote_type="both"):
        if quote_type not in ["both", "single", "double"]:
            raise ValueError(
                f"{quote_type} is not a valid quote_type. Specify 'both', 'single', or 'double'",
            )
        self.quote_type = quote_type
        IN_DOUBLE_QUOTES = r'((^|\W)"(.)*?"(?!\w))'
        IN_SINGLE_QUOTES = r"((^|\W)'(.)*?'(?!\w))"
        if quote_type == "double":
            self.regex = IN_DOUBLE_QUOTES
        elif quote_type == "single":
            self.regex = IN_SINGLE_QUOTES
        else:
            self.regex = f"({IN_SINGLE_QUOTES}|{IN_DOUBLE_QUOTES})"

    def get_function(self):
        def count_words_in_quotes(text):
            if pd.isnull(text):
                return pd.NA
            matches = re.findall(self.regex, text, re.DOTALL)
            count = 0
            for match in matches:
                matched_phrase = match[0]
                words = re.split(f"{DELIMITERS}", matched_phrase)
                for word in words:
                    if len(word.strip(punctuation + " ")):
                        count += 1
            return count

        def num_words_in_quotes(array):
            return array.apply(count_words_in_quotes).astype("Int64")

        return num_words_in_quotes



import pandas as pd
from scipy.signal import find_peaks
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Integer

from featuretools.primitives.base import AggregationPrimitive


class NumPeaks(AggregationPrimitive):
    """Determines the number of peaks in a list of numbers.

    Description:
        Given a list of numbers, count the number of local
        maxima. Uses the find_peaks function from scipy.signal.

    Examples:
        >>> num_peaks = NumPeaks()
        >>> num_peaks([-5, 0, 10, 0, 10, -5, -4, -5, 10, 0])
        4
    """

    name = "num_peaks"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def num_peaks(x):
            if x.dtype == "Int64":
                x = x.astype("float64")
            peaks = find_peaks(x)[0]
            return len(peaks[~pd.isna(peaks)])

        return num_peaks



import numpy as np
import pandas as pd

from featuretools.primitives import NumUniqueSeparators
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestNumUniqueSeparators(PrimitiveTestBase):
    primitive = NumUniqueSeparators

    def test_punctuation(self):
        x = pd.Series(
            [
                "This: is a test file",
                "This, is second line?",
                "third/line $1,000;",
                "and--subsequen't lines...",
                "*and, more..",
            ],
        )
        primitive_func = self.primitive().get_function()
        answers = pd.Series([1, 3, 3, 2, 3])
        pd.testing.assert_series_equal(primitive_func(x), answers, check_names=False)

    def test_other_delimeters(self):
        x = pd.Series(["@#$%^&*()<>/[]\\`~-_=+"])
        primitive_func = self.primitive().get_function()
        answers = pd.Series([0])
        pd.testing.assert_series_equal(primitive_func(x), answers, check_names=False)

    def test_multiline(self):
        x = pd.Series(
            [
                "This is a test file",
                "This is second line\nthird line $1000;\nand subsequent lines",
                "and more!",
            ],
        )
        primitive_func = self.primitive().get_function()
        answers = pd.Series([1, 3, 2])
        pd.testing.assert_series_equal(primitive_func(x), answers, check_names=False)

    def test_nans(self):
        x = pd.Series([np.nan, "", "third line."])
        primitive_func = self.primitive().get_function()
        answers = pd.Series([pd.NA, 0, 2])
        pd.testing.assert_series_equal(primitive_func(x), answers, check_names=False)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



from featuretools.primitives.standard.aggregation.all_primitive import All
from featuretools.primitives.standard.aggregation.any_primitive import Any
from featuretools.primitives.standard.aggregation.avg_time_between import AvgTimeBetween
from featuretools.primitives.standard.aggregation.average_count_per_unique import (
    AverageCountPerUnique,
)
from featuretools.primitives.standard.aggregation.count import Count
from featuretools.primitives.standard.aggregation.count_above_mean import CountAboveMean
from featuretools.primitives.standard.aggregation.count_below_mean import CountBelowMean
from featuretools.primitives.standard.aggregation.count_greater_than import (
    CountGreaterThan,
)
from featuretools.primitives.standard.aggregation.count_inside_nth_std import (
    CountInsideNthSTD,
)
from featuretools.primitives.standard.aggregation.count_inside_range import (
    CountInsideRange,
)
from featuretools.primitives.standard.aggregation.count_less_than import CountLessThan
from featuretools.primitives.standard.aggregation.count_outside_nth_std import (
    CountOutsideNthSTD,
)
from featuretools.primitives.standard.aggregation.count_outside_range import (
    CountOutsideRange,
)
from featuretools.primitives.standard.aggregation.date_first_event import DateFirstEvent
from featuretools.primitives.standard.aggregation.entropy import Entropy
from featuretools.primitives.standard.aggregation.first import First
from featuretools.primitives.standard.aggregation.first_last_time_delta import (
    FirstLastTimeDelta,
)
from featuretools.primitives.standard.aggregation.kurtosis import Kurtosis
from featuretools.primitives.standard.aggregation.is_unique import IsUnique
from featuretools.primitives.standard.aggregation.last import Last
from featuretools.primitives.standard.aggregation.max_primitive import Max
from featuretools.primitives.standard.aggregation.max_consecutive_false import (
    MaxConsecutiveFalse,
)
from featuretools.primitives.standard.aggregation.max_consecutive_negatives import (
    MaxConsecutiveNegatives,
)
from featuretools.primitives.standard.aggregation.max_consecutive_positives import (
    MaxConsecutivePositives,
)
from featuretools.primitives.standard.aggregation.max_consecutive_true import (
    MaxConsecutiveTrue,
)
from featuretools.primitives.standard.aggregation.max_consecutive_zeros import (
    MaxConsecutiveZeros,
)
from featuretools.primitives.standard.aggregation.mean import Mean
from featuretools.primitives.standard.aggregation.median import Median
from featuretools.primitives.standard.aggregation.max_count import MaxCount
from featuretools.primitives.standard.aggregation.median_count import MedianCount
from featuretools.primitives.standard.aggregation.max_min_delta import MaxMinDelta
from featuretools.primitives.standard.aggregation.min_count import MinCount
from featuretools.primitives.standard.aggregation.min_primitive import Min
from featuretools.primitives.standard.aggregation.mode import Mode
from featuretools.primitives.standard.aggregation.n_unique_days import NUniqueDays
from featuretools.primitives.standard.aggregation.n_unique_days_of_calendar_year import (
    NUniqueDaysOfCalendarYear,
)
from featuretools.primitives.standard.aggregation.n_unique_days_of_month import (
    NUniqueDaysOfMonth,
)
from featuretools.primitives.standard.aggregation.has_no_duplicates import (
    HasNoDuplicates,
)
from featuretools.primitives.standard.aggregation.is_monotonically_decreasing import (
    IsMonotonicallyDecreasing,
)
from featuretools.primitives.standard.aggregation.is_monotonically_increasing import (
    IsMonotonicallyIncreasing,
)
from featuretools.primitives.standard.aggregation.n_unique_months import NUniqueMonths
from featuretools.primitives.standard.aggregation.n_unique_weeks import NUniqueWeeks
from featuretools.primitives.standard.aggregation.n_most_common import NMostCommon
from featuretools.primitives.standard.aggregation.n_most_common_frequency import (
    NMostCommonFrequency,
)
from featuretools.primitives.standard.aggregation.num_true import NumTrue
from featuretools.primitives.standard.aggregation.num_peaks import NumPeaks
from featuretools.primitives.standard.aggregation.num_zero_crossings import (
    NumZeroCrossings,
)
from featuretools.primitives.standard.aggregation.num_true_since_last_false import (
    NumTrueSinceLastFalse,
)
from featuretools.primitives.standard.aggregation.num_false_since_last_true import (
    NumFalseSinceLastTrue,
)
from featuretools.primitives.standard.aggregation.num_consecutive_greater_mean import (
    NumConsecutiveGreaterMean,
)
from featuretools.primitives.standard.aggregation.num_consecutive_less_mean import (
    NumConsecutiveLessMean,
)
from featuretools.primitives.standard.aggregation.num_unique import NumUnique
from featuretools.primitives.standard.aggregation.percent_unique import PercentUnique
from featuretools.primitives.standard.aggregation.percent_true import PercentTrue
from featuretools.primitives.standard.aggregation.skew import Skew
from featuretools.primitives.standard.aggregation.std import Std
from featuretools.primitives.standard.aggregation.sum_primitive import Sum
from featuretools.primitives.standard.aggregation.time_since_first import TimeSinceFirst
from featuretools.primitives.standard.aggregation.time_since_last import TimeSinceLast
from featuretools.primitives.standard.aggregation.time_since_last_true import (
    TimeSinceLastTrue,
)
from featuretools.primitives.standard.aggregation.time_since_last_min import (
    TimeSinceLastMin,
)
from featuretools.primitives.standard.aggregation.time_since_last_max import (
    TimeSinceLastMax,
)
from featuretools.primitives.standard.aggregation.time_since_last_false import (
    TimeSinceLastFalse,
)
from featuretools.primitives.standard.aggregation.trend import Trend
from featuretools.primitives.standard.aggregation.variance import Variance
